# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Joint Policy-Space Response Oracles.

An implementation of JSPRO, described in https://arxiv.org/abs/2106.09435.

Bibtex / Cite:

```
@misc{marris2021multiagent,
    title={Multi-Agent Training beyond Zero-Sum with Correlated Equilibrium
           Meta-Solvers},
    author={Luke Marris and Paul Muller and Marc Lanctot and Karl Tuyls and
            Thore Graepel},
    year={2021},
    eprint={2106.09435},
    archivePrefix={arXiv},
    primaryClass={cs.MA}
}
```
"""

import itertools
import string

from absl import logging

import cvxpy as cp
import numpy as np
import scipy as sp

from open_spiel.python import policy
from open_spiel.python.algorithms import projected_replicator_dynamics
from open_spiel.python.egt import alpharank as alpharank_lib
import pyspiel


DEFAULT_ECOS_SOLVER_KWARGS = dict(
    solver="ECOS",
    max_iters=100000000,
    abstol=1e-7,
    reltol=1e-7,
    feastol=1e-7,
    abstol_inacc=1e-7,
    reltol_inacc=1e-7,
    feastol_inacc=1e-7,
    verbose=False,
)
DEFAULT_OSQP_SOLVER_KWARGS = dict(
    solver="OSQP",
    max_iter=1000000000,
    eps_abs=1e-8,
    eps_rel=1e-8,
    eps_prim_inf=1e-8,
    eps_dual_inf=1e-8,
    polish_refine_iter=100,
    check_termination=1000,
    sigma=1e-7,  # Default 1e-6
    delta=1e-7,  # Default 1e-06
    verbose=False,
)
DEFAULT_CVXOPT_SOLVER_KWARGS = dict(
    solver="CVXOPT",
    maxiters=200000,
    abstol=5e-8,
    reltol=5e-8,
    feastol=5e-8,
    refinement=10,
    verbose=False,
)
INIT_POLICIES = (
    "uniform",  # Unopinionated but slower to evaluate.
    "random_deterministic",  # Faster to evaluate but requires samples.
)
UPDATE_PLAYERS_STRATEGY = (
    "all",
    "cycle",
    "random",
)
BRS = (
    "cce",
    "ce",
)
BR_SELECTIONS = (
    "all",  # All policies.
    "all_novel",  # All novel policies.
    "random",  # Random.
    "random_novel",  # Random novel BR (one that has not be considered before).
    "largest_gap",  # The BR with the largest gap.
)
META_SOLVERS = (
    "uni",  # Uniform.
    "undominated_uni",  # Uniform over undominated strategies.
    "rj",  # Random joint.
    "undominated_rj",  # Random joint.
    "rd",  # Random dirichlet.
    "undominated_rd",  # Random dirichlet.
    "prd",  # Prodected replicator dynamics.
    "alpharank",  # Alpha-Rank
    "mgce",  # Maximum gini CE.
    "min_epsilon_mgce",  # Min Epsilon Maximum gini CE.
    "approx_mgce",  # Approximate Maximum gini CE.
    "rmwce",  # Random maximum welfare CE.
    "mwce",  # Maximum welfare CE.
    "rvce",  # Random vertex CE.
    "mgcce",  # Maximum gini CCE.
    "min_epsilon_mgcce",  # Min Epsilon Maximum gini CCE.
    "approx_mgcce",  # Approximate Maximum gini CE.
    "rmwcce",  # Random maximum welfare CCE.
    "mwcce",  # Maximum welfare CCE.
    "rvcce",  # Random vertex CCE.
)
LOG_STRING = """
Iteration {iteration: 6d}
=== ({game})
Player            {player}
BRs               {brs}
Num Policies      {num_policies}
Unique Policies   {unique}
--- ({train_meta_solver})
Train Value       {train_value}
Train Gap         {train_gap}
--- ({eval_meta_solver})
Eval Value        {eval_value}
Eval Gap          {eval_gap}
"""
DIST_TOL = 1e-8
GAP_TOL = 1e-8
RETURN_TOL = 1e-12


## Meta Solvers.


# Helper Functions - Dominated strategy elimination.
def _eliminate_dominated_payoff(
    payoff, epsilon, action_labels=None, action_repeats=None, weakly=False):
  """Eliminate epsilon dominated strategies."""
  num_players = payoff.shape[0]
  eliminated = True
  if action_labels is None:
    action_labels = [np.arange(na, dtype=np.int32) for na in payoff.shape[1:]]
  if action_repeats is not None:
    action_repeats = [ar for ar in action_repeats]
  while eliminated:
    eliminated = False
    for p in range(num_players):
      if epsilon > 0.0:
        continue
      num_actions = payoff.shape[1:]
      if num_actions[p] <= 1:
        continue
      for a in range(num_actions[p]):
        index = [slice(None) for _ in range(num_players)]
        index[p] = slice(a, a+1)
        if weakly:
          diff = payoff[p] <= payoff[p][tuple(index)]
        else:
          diff = payoff[p] < payoff[p][tuple(index)]
        axis = tuple(range(p)) + tuple(range(p+1, num_players))
        less = np.all(diff, axis=axis)
        less[a] = False  # Action cannot eliminate itself.
        if np.any(less):
          nonzero = np.nonzero(less)
          payoff = np.delete(payoff, nonzero, axis=p+1)
          action_labels[p] = np.delete(action_labels[p], nonzero)
          if action_repeats is not None:
            action_repeats[p] = np.delete(action_repeats[p], nonzero)
          eliminated = True
          break
  return payoff, action_labels, action_repeats


def _reconstruct_dist(eliminated_dist, action_labels, num_actions):
  """Returns reconstructed dist from eliminated_dist and action_labels.

  Redundant dist elements are given values 0.

  Args:
    eliminated_dist: Array of shape [A0E, A1E, ...].
    action_labels: List of length N and shapes [[A0E], [A1E], ...].
    num_actions: List of length N and values [A0, A1, ...].

  Returns:
    reconstructed_dist: Array of shape [A0, A1, ...].
  """
  reconstructed_payoff = np.zeros(num_actions)
  reconstructed_payoff[np.ix_(*action_labels)] = eliminated_dist
  return reconstructed_payoff


def _eliminate_dominated_decorator(func):
  """Wrap eliminate dominated."""
  def wrapper(payoff, per_player_repeats, *args, eliminate_dominated=True,
              **kwargs):
    epsilon = getattr(kwargs, "epsilon", 0.0)
    if not eliminate_dominated:
      return func(payoff, *args, **kwargs)
    num_actions = payoff.shape[1:]
    eliminated_payoff, action_labels, eliminated_action_repeats = _eliminate_dominated_payoff(
        payoff, epsilon, action_repeats=per_player_repeats)
    eliminated_dist, meta = func(
        eliminated_payoff, eliminated_action_repeats, *args, **kwargs)
    meta["eliminated_dominated_dist"] = eliminated_dist
    meta["eliminated_dominated_payoff"] = eliminated_payoff
    dist = _reconstruct_dist(
        eliminated_dist, action_labels, num_actions)
    return dist, meta
  return wrapper


# Optimization.
def _try_two_solvers(func, *args, **kwargs):
  try:
    logging.debug("Trying CVXOPT.", flush=True)
    kwargs_ = {"solver_kwargs": DEFAULT_CVXOPT_SOLVER_KWARGS, **kwargs}
    res = func(*args, **kwargs_)
  except:  # pylint: disable=bare-except
    logging.debug("CVXOPT failed. Trying OSQP.", flush=True)
    kwargs_ = {"solver_kwargs": DEFAULT_OSQP_SOLVER_KWARGS, **kwargs}
    res = func(*args, **kwargs_)
  return res


# Helper Functions - CCEs.
def _indices(p, a, num_players):
  return [a if p_ == p else slice(None) for p_ in range(num_players)]


def _sparse_indices_generator(player, action, num_actions):
  indices = [(action,) if p == player else range(na)
             for p, na in enumerate(num_actions)]
  return itertools.product(*indices)


def _partition_by_player(val, p_vec, num_players):
  """Partitions a value by the players vector."""
  parts = []
  for p in range(num_players):
    inds = p_vec == p
    if inds.size > 0:
      parts.append(val[inds])
    else:
      parts.append(None)
  return parts


def _cce_constraints(payoff, epsilons, remove_null=True, zero_tolerance=1e-8):
  """Returns the coarse correlated constraints.

  Args:
    payoff: A [NUM_PLAYER, NUM_ACT_0, NUM_ACT_1, ...] shape payoff tensor.
    epsilons: Per player floats corresponding to the epsilon.
    remove_null: Remove null rows of the constraint matrix.
    zero_tolerance: Zero out elements with small value.

  Returns:
    a_mat: The gain matrix for deviting to an action or shape [SUM(A), PROD(A)].
    meta: Dictionary containing meta information.
  """
  num_players = payoff.shape[0]
  num_actions = payoff.shape[1:]
  num_dists = int(np.prod(num_actions))

  cor_cons = int(np.sum(num_actions))

  a_mat = np.zeros([cor_cons] + list(num_actions))
  p_vec = np.zeros([cor_cons], dtype=np.int32)
  i_vec = np.zeros([cor_cons], dtype=np.int32)
  con = 0
  for p in range(num_players):
    for a1 in range(num_actions[p]):
      a1_inds = tuple(_indices(p, a1, num_players))
      for a0 in range(num_actions[p]):
        a0_inds = tuple(_indices(p, a0, num_players))
        a_mat[con][a0_inds] += payoff[p][a1_inds]
      a_mat[con] -= payoff[p]
      a_mat[con] -= epsilons[p]

      p_vec[con] = p
      i_vec[con] = a0

      con += 1

  a_mat = np.reshape(a_mat, [cor_cons, num_dists])
  a_mat[np.abs(a_mat) < zero_tolerance] = 0.0
  if remove_null:
    null_cons = np.any(a_mat != 0.0, axis=-1)
    redundant_cons = np.max(a_mat, axis=1) >= 0
    nonzero_mask = null_cons & redundant_cons
    a_mat = a_mat[nonzero_mask, :].copy()
    p_vec = p_vec[nonzero_mask].copy()
    i_vec = i_vec[nonzero_mask].copy()

  meta = dict(
      p_vec=p_vec,
      i_vec=i_vec,
      epsilons=epsilons,
  )

  return a_mat, meta


def _ace_constraints(payoff, epsilons, remove_null=True, zero_tolerance=0.0):
  """Returns sparse alternate ce constraints Ax - epsilon <= 0.

  Args:
    payoff: Dense payoff tensor.
    epsilons: Scalar epsilon approximation.
    remove_null: Whether to remove null row constraints.
    zero_tolerance: Smallest absolute value.

  Returns:
    a_csr: Sparse gain matrix from switching from one action to another.
    e_vec: Epsilon vector.
    meta: Dictionary containing meta information.
  """
  num_players = payoff.shape[0]
  num_actions = payoff.shape[1:]
  num_dists = int(np.prod(num_actions))

  num_cons = 0
  for p in range(num_players):
    num_cons += num_actions[p] * (num_actions[p] - 1)

  a_dok = sp.sparse.dok_matrix((num_cons, num_dists))
  e_vec = np.zeros([num_cons])
  p_vec = np.zeros([num_cons], dtype=np.int32)
  i_vec = np.zeros([num_cons, 2], dtype=np.int32)

  num_null_cons = None
  num_redundant_cons = None
  num_removed_cons = None

  if num_cons > 0:
    con = 0
    for p in range(num_players):
      generator = itertools.permutations(range(num_actions[p]), 2)
      for a0, a1 in generator:
        a0_inds = _sparse_indices_generator(p, a0, num_actions)
        a1_inds = _sparse_indices_generator(p, a1, num_actions)

        for a0_ind, a1_ind in zip(a0_inds, a1_inds):
          a0_ind_flat = np.ravel_multi_index(a0_ind, num_actions)
          val = payoff[p][a1_ind] - payoff[p][a0_ind]
          if abs(val) > zero_tolerance:
            a_dok[con, a0_ind_flat] = val

        e_vec[con] = epsilons[p]
        p_vec[con] = p
        i_vec[con] = [a0, a1]
        con += 1

    a_csr = a_dok.tocsr()
    if remove_null:
      null_cons = np.logical_or(
          a_csr.max(axis=1).todense() != 0.0,
          a_csr.min(axis=1).todense() != 0.0)
      null_cons = np.ravel(null_cons)
      redundant_cons = np.ravel(a_csr.max(axis=1).todense()) >= e_vec
      nonzero_mask = null_cons & redundant_cons
      a_csr = a_csr[nonzero_mask, :]
      e_vec = e_vec[nonzero_mask].copy()
      p_vec = p_vec[nonzero_mask].copy()
      i_vec = i_vec[nonzero_mask].copy()
      num_null_cons = np.sum(~null_cons)
      num_redundant_cons = np.sum(~redundant_cons)
      num_removed_cons = np.sum(~nonzero_mask)

  else:
    a_csr = a_dok.tocsr()

  meta = dict(
      p_vec=p_vec,
      i_vec=i_vec,
      epsilons=epsilons,
      num_null_cons=num_null_cons,
      num_redundant_cons=num_redundant_cons,
      num_removed_cons=num_removed_cons,
  )

  return a_csr, e_vec, meta


def _get_repeat_factor(action_repeats):
  """Returns the repeat factors for the game."""
  num_players = len(action_repeats)
  out_labels = string.ascii_lowercase[:len(action_repeats)]
  in_labels = ",".join(out_labels)
  repeat_factor = np.ravel(np.einsum(
      "{}->{}".format(in_labels, out_labels), *action_repeats))
  indiv_repeat_factors = []
  for player in range(num_players):
    action_repeats_ = [
        np.ones_like(ar) if player == p else ar
        for p, ar in enumerate(action_repeats)]
    indiv_repeat_factor = np.ravel(np.einsum(
        "{}->{}".format(in_labels, out_labels), *action_repeats_))
    indiv_repeat_factors.append(indiv_repeat_factor)
  return repeat_factor, indiv_repeat_factors


# Solvers.
def _linear(
    payoff,
    a_mat,
    e_vec,
    action_repeats=None,
    solver_kwargs=None,
    cost=None):
  """Returns linear solution.

  This is a linear program.

  Args:
    payoff: A [NUM_PLAYER, NUM_ACT_0, NUM_ACT_1, ...] shape payoff tensor.
    a_mat: Constaint matrix.
    e_vec: Epsilon vector.
    action_repeats: List of action repeat counts.
    solver_kwargs: Solver kwargs.
    cost: Cost function of same shape as payoff.

  Returns:
    An epsilon-correlated equilibrium.
  """
  num_players = payoff.shape[0]
  num_actions = payoff.shape[1:]
  num_dists = int(np.prod(num_actions))

  if solver_kwargs is None:
    solver_kwargs = DEFAULT_ECOS_SOLVER_KWARGS

  if a_mat.shape[0] > 0:
    # Variables.
    x = cp.Variable(num_dists, nonneg=True)

    # Classifier.
    epsilon_dists = cp.matmul(a_mat, x) - e_vec

    # Constraints.
    dist_eq_con = cp.sum(x) == 1
    cor_lb_con = epsilon_dists <= 0

    # Objective.
    if cost is None:
      player_totals = [
          cp.sum(cp.multiply(payoff[p].flat, x)) for p in range(num_players)]
      reward = cp.sum(player_totals)
    else:
      reward = cp.sum(cp.multiply(cost.flat, x))
    obj = cp.Maximize(reward)

    prob = cp.Problem(obj, [
        dist_eq_con,
        cor_lb_con,
    ])

    # Solve.
    prob.solve(**solver_kwargs)
    status = prob.status

    # Distribution.
    dist = np.reshape(x.value, num_actions)

    # Other.
    val = reward.value
  else:
    if action_repeats is not None:
      repeat_factor, _ = _get_repeat_factor(action_repeats)
      x = repeat_factor / np.sum(repeat_factor)
    else:
      x = np.ones([num_dists]) / num_dists
    val = 0.0  # Fix me.
    dist = np.reshape(x, num_actions)
    status = None

  meta = dict(
      x=x,
      a_mat=a_mat,
      val=val,
      status=status,
      payoff=payoff,
      consistent=True,
      unique=False,
  )

  return dist, meta


def _qp_cce(
    payoff,
    a_mats,
    e_vecs,
    assume_full_support=False,
    action_repeats=None,
    solver_kwargs=None,
    min_epsilon=False):
  """Returns the correlated equilibrium with maximum Gini impurity.

  Args:
    payoff: A [NUM_PLAYER, NUM_ACT_0, NUM_ACT_1, ...] shape payoff tensor.
    a_mats: A [NUM_CON, PROD(A)] shape gain tensor.
    e_vecs: Epsilon vector.
    assume_full_support: Whether to ignore beta values.
    action_repeats: Vector of action repeats for each player.
    solver_kwargs: Additional kwargs for solver.
    min_epsilon: Whether to minimize epsilon.

  Returns:
    An epsilon-correlated equilibrium.
  """
  num_players = payoff.shape[0]
  num_actions = payoff.shape[1:]
  num_dists = int(np.prod(num_actions))

  if solver_kwargs is None:
    solver_kwargs = DEFAULT_OSQP_SOLVER_KWARGS

  epsilon = None
  nonzero_cons = [a_mat.shape[0] > 0 for a_mat in a_mats if a_mat is not None]
  if any(nonzero_cons):
    x = cp.Variable(num_dists, nonneg=(not assume_full_support))
    if min_epsilon:
      epsilon = cp.Variable(nonpos=True)
      e_vecs = [epsilon] * num_players

    if action_repeats is not None:
      repeat_factor, _ = _get_repeat_factor(action_repeats)
      x_repeated = cp.multiply(x, repeat_factor)
      dist_eq_con = cp.sum(x_repeated) == 1
      cor_lb_cons = [
          cp.matmul(a_mat, cp.multiply(x, repeat_factor)) <= e_vec
          for a_mat, e_vec in
          zip(a_mats, e_vecs) if a_mat.size > 0]
      eye = sp.sparse.diags(repeat_factor)
    else:
      repeat_factor = 1
      x_repeated = x
      dist_eq_con = cp.sum(x_repeated) == 1
      cor_lb_cons = [
          cp.matmul(a_mat, x) <= e_vec for a_mat, e_vec in
          zip(a_mats, e_vecs) if a_mat.size > 0]
      eye = sp.sparse.eye(num_dists)

    # This is more memory efficient than using cp.sum_squares.
    cost = 1 - cp.quad_form(x, eye)
    if min_epsilon:
      cost -= cp.multiply(2, epsilon)

    obj = cp.Maximize(cost)
    prob = cp.Problem(obj, [dist_eq_con] + cor_lb_cons)
    cost_value = prob.solve(**solver_kwargs)
    status = prob.status
    alphas = [cor_lb_con.dual_value for cor_lb_con in cor_lb_cons]
    lamb = dist_eq_con.dual_value

    val = cost.value
    x = x_repeated.value
    dist = np.reshape(x, num_actions)
  else:
    cost_value = 0.0
    val = 1 - 1 / num_dists
    if action_repeats is not None:
      repeat_factor, _ = _get_repeat_factor(action_repeats)
      x = repeat_factor / np.sum(repeat_factor)
    else:
      x = np.ones([num_dists]) / num_dists
    dist = np.reshape(x, num_actions)
    status = None
    alphas = [np.zeros([])]
    lamb = None

  meta = dict(
      x=x,
      a_mats=a_mats,
      status=status,
      cost=cost_value,
      val=val,
      alphas=alphas,
      lamb=lamb,
      unique=True,
      min_epsilon=None if epsilon is None else epsilon.value,
  )
  return dist, meta


def _qp_ce(
    payoff,
    a_mats,
    e_vecs,
    assume_full_support=False,
    action_repeats=None,
    solver_kwargs=None,
    min_epsilon=False):
  """Returns the correlated equilibrium with maximum Gini impurity.

  Args:
    payoff: A [NUM_PLAYER, NUM_ACT_0, NUM_ACT_1, ...] shape payoff tensor.
    a_mats: A [NUM_CON, PROD(A)] shape gain tensor.
    e_vecs: Epsilon vector.
    assume_full_support: Whether to ignore beta values.
    action_repeats: Vector of action repeats for each player.
    solver_kwargs: Additional kwargs for solver.
    min_epsilon: Whether to minimize epsilon.

  Returns:
    An epsilon-correlated equilibrium.
  """
  num_players = payoff.shape[0]
  num_actions = payoff.shape[1:]
  num_dists = int(np.prod(num_actions))

  if solver_kwargs is None:
    solver_kwargs = DEFAULT_OSQP_SOLVER_KWARGS

  epsilon = None
  nonzero_cons = [a_mat.shape[0] > 0 for a_mat in a_mats if a_mat is not None]
  if any(nonzero_cons):
    x = cp.Variable(num_dists, nonneg=(not assume_full_support))
    if min_epsilon:
      epsilon = cp.Variable(nonpos=True)
      e_vecs = [epsilon] * num_players

    if action_repeats is not None:
      repeat_factor, indiv_repeat_factors = _get_repeat_factor(
          action_repeats)
      x_repeated = cp.multiply(x, repeat_factor)
      dist_eq_con = cp.sum(x_repeated) == 1
      cor_lb_cons = [
          cp.matmul(a_mat, cp.multiply(x, rf)) <= e_vec for a_mat, e_vec, rf in
          zip(a_mats, e_vecs, indiv_repeat_factors) if a_mat.size > 0]
      eye = sp.sparse.diags(repeat_factor)
    else:
      repeat_factor = 1
      x_repeated = x
      dist_eq_con = cp.sum(x_repeated) == 1
      cor_lb_cons = [
          cp.matmul(a_mat, x) <= e_vec for a_mat, e_vec in
          zip(a_mats, e_vecs) if a_mat.size > 0]
      eye = sp.sparse.eye(num_dists)

    # This is more memory efficient than using cp.sum_squares.
    cost = 1 - cp.quad_form(x, eye)
    if min_epsilon:
      cost -= cp.multiply(2, epsilon)

    obj = cp.Maximize(cost)
    prob = cp.Problem(obj, [dist_eq_con] + cor_lb_cons)
    cost_value = prob.solve(**solver_kwargs)
    status = prob.status
    alphas = [cor_lb_con.dual_value for cor_lb_con in cor_lb_cons]
    lamb = dist_eq_con.dual_value

    val = cost.value
    x = x_repeated.value
    dist = np.reshape(x, num_actions)
  else:
    cost_value = 0.0
    val = 1 - 1 / num_dists
    if action_repeats is not None:
      repeat_factor, indiv_repeat_factors = _get_repeat_factor(
          action_repeats)
      x = repeat_factor / np.sum(repeat_factor)
    else:
      x = np.ones([num_dists]) / num_dists
    dist = np.reshape(x, num_actions)
    status = None
    alphas = [np.zeros([])]
    lamb = None

  meta = dict(
      x=x,
      a_mats=a_mats,
      status=status,
      cost=cost_value,
      val=val,
      alphas=alphas,
      lamb=lamb,
      unique=True,
      min_epsilon=None if epsilon is None else epsilon.value,
  )
  return dist, meta


def _expand_meta_game(meta_game, per_player_repeats):
  num_players = meta_game.shape[0]
  for player in range(num_players):
    meta_game = np.repeat(meta_game, per_player_repeats[player], axis=player+1)
  return meta_game


def _unexpand_meta_dist(meta_dist, per_player_repeats):
  num_players = len(meta_dist.shape)
  for player in range(num_players):
    meta_dist = np.add.reduceat(
        meta_dist, [0] + np.cumsum(per_player_repeats[player]).tolist()[:-1],
        axis=player)
  return meta_dist


# Meta-solvers - Baselines.
def _uni(meta_game, per_player_repeats, ignore_repeats=False):
  """Uniform."""
  if ignore_repeats:
    num_policies = meta_game.shape[1:]
    num_dists = np.prod(num_policies)
    meta_dist = np.full(num_policies, 1./num_dists)
  else:
    outs = [ppr / np.sum(ppr) for ppr in per_player_repeats]
    labels = string.ascii_lowercase[:len(outs)]
    comma_labels = ",".join(labels)
    meta_dist = np.einsum("{}->{}".format(comma_labels, labels), *outs)
  return meta_dist, dict()


@_eliminate_dominated_decorator
def _undominated_uni(meta_game, per_player_repeats, ignore_repeats=False):
  """Undominated uniform."""
  return _uni(meta_game, per_player_repeats, ignore_repeats=ignore_repeats)


def _rj(meta_game, per_player_repeats, ignore_repeats=False):
  """Random joint."""
  ignore_repeats = True
  pvals, _ = _uni(
      meta_game, per_player_repeats, ignore_repeats=ignore_repeats)
  meta_dist = np.reshape(
      np.random.multinomial(1, pvals.flat), pvals.shape).astype(np.float64)
  return meta_dist, dict()


@_eliminate_dominated_decorator
def _undominated_rj(meta_game, per_player_repeats, ignore_repeats=False):
  """Undominated random joint."""
  return _rj(meta_game, per_player_repeats, ignore_repeats=ignore_repeats)


def _rd(meta_game, per_player_repeats, ignore_repeats=False):
  """Random dirichlet."""
  ignore_repeats = True
  if ignore_repeats:
    num_policies = meta_game.shape[1:]
    alpha = np.ones(num_policies)
  else:
    outs = [ppr for ppr in per_player_repeats]
    labels = string.ascii_lowercase[:len(outs)]
    comma_labels = ",".join(labels)
    alpha = np.einsum("{}->{}".format(comma_labels, labels), *outs)
  meta_dist = np.reshape(
      np.random.dirichlet(alpha.flat), alpha.shape).astype(np.float64)
  return meta_dist, dict()


@_eliminate_dominated_decorator
def _undominated_rd(meta_game, per_player_repeats, ignore_repeats=False):
  """Undominated random dirichlet."""
  return _rd(meta_game, per_player_repeats, ignore_repeats=ignore_repeats)


def _prd(meta_game, per_player_repeats, ignore_repeats=False):
  """Projected replicator dynamics."""
  if not ignore_repeats:
    meta_game = _expand_meta_game(meta_game, per_player_repeats)
  meta_dist = projected_replicator_dynamics.projected_replicator_dynamics(
      meta_game)
  labels = string.ascii_lowercase[:len(meta_dist)]
  comma_labels = ",".join(labels)
  meta_dist = np.einsum("{}->{}".format(comma_labels, labels), *meta_dist)
  meta_dist[meta_dist < DIST_TOL] = 0.0
  meta_dist /= np.sum(meta_dist)
  meta_dist = _unexpand_meta_dist(meta_dist, per_player_repeats)
  return meta_dist, dict()


@_eliminate_dominated_decorator
def _alpharank(meta_game, per_player_repeats, ignore_repeats=False):
  """AlphaRank."""
  if not ignore_repeats:
    meta_game = _expand_meta_game(meta_game, per_player_repeats)
  meta_dist = alpharank_lib.sweep_pi_vs_epsilon([mg for mg in meta_game])
  meta_dist[meta_dist < DIST_TOL] = 0.0
  meta_dist /= np.sum(meta_dist)
  meta_dist = np.reshape(meta_dist, meta_game.shape[1:])
  if not ignore_repeats:
    meta_dist = _unexpand_meta_dist(meta_dist, per_player_repeats)
  return meta_dist, dict()


# Meta-solvers - CEs.
@_eliminate_dominated_decorator
def _mgce(meta_game, per_player_repeats, ignore_repeats=False):
  """Maximum Gini CE."""
  a_mat, e_vec, meta = _ace_constraints(
      meta_game, [0.0] * len(per_player_repeats), remove_null=True,
      zero_tolerance=1e-8)
  a_mats = _partition_by_player(
      a_mat, meta["p_vec"], len(per_player_repeats))
  e_vecs = _partition_by_player(
      e_vec, meta["p_vec"], len(per_player_repeats))
  dist, _ = _try_two_solvers(
      _qp_ce,
      meta_game, a_mats, e_vecs,
      action_repeats=(None if ignore_repeats else per_player_repeats))
  return dist, dict()


@_eliminate_dominated_decorator
def _min_epsilon_mgce(meta_game, per_player_repeats, ignore_repeats=False):
  """Min Epsilon Maximum Gini CE."""
  a_mat, e_vec, meta = _ace_constraints(
      meta_game, [0.0] * len(per_player_repeats), remove_null=True,
      zero_tolerance=1e-8)
  a_mats = _partition_by_player(
      a_mat, meta["p_vec"], len(per_player_repeats))
  e_vecs = _partition_by_player(
      e_vec, meta["p_vec"], len(per_player_repeats))
  dist, _ = _try_two_solvers(
      _qp_ce,
      meta_game, a_mats, e_vecs,
      action_repeats=(None if ignore_repeats else per_player_repeats),
      min_epsilon=True)
  return dist, dict()


@_eliminate_dominated_decorator
def _approx_mgce(meta_game, per_player_repeats, ignore_repeats=False,
                 epsilon=0.01):
  """Approximate Maximum Gini CE."""
  a_mat, e_vec, meta = _ace_constraints(
      meta_game, [0.0] * len(per_player_repeats), remove_null=True,
      zero_tolerance=1e-8)
  max_ab = 0.0
  if a_mat.size:
    max_ab = np.max(a_mat.mean(axis=1))
  a_mat, e_vec, meta = _ace_constraints(
      meta_game, [epsilon * max_ab] * len(per_player_repeats), remove_null=True,
      zero_tolerance=1e-8)
  a_mats = _partition_by_player(
      a_mat, meta["p_vec"], len(per_player_repeats))
  e_vecs = _partition_by_player(
      e_vec, meta["p_vec"], len(per_player_repeats))
  dist, _ = _try_two_solvers(
      _qp_ce,
      meta_game, a_mats, e_vecs,
      action_repeats=(None if ignore_repeats else per_player_repeats))
  return dist, dict()


@_eliminate_dominated_decorator
def _rmwce(meta_game, per_player_repeats, ignore_repeats=False):
  """Random maximum welfare CE."""
  del ignore_repeats
  num_players = len(per_player_repeats)
  cost = np.ravel(np.sum(meta_game, axis=0))
  cost += np.ravel(np.random.normal(size=meta_game.shape[1:])) * 1e-6
  a_mat, e_vec, _ = _ace_constraints(
      meta_game, [0.0] * num_players, remove_null=True,
      zero_tolerance=1e-8)
  x, _ = _linear(meta_game, a_mat, e_vec, cost=cost)
  dist = np.reshape(x, meta_game.shape[1:])
  return dist, dict()


@_eliminate_dominated_decorator
def _mwce(meta_game, per_player_repeats, ignore_repeats=False):
  """Maximum welfare CE."""
  del ignore_repeats
  num_players = len(per_player_repeats)
  cost = np.ravel(np.sum(meta_game, axis=0))
  a_mat, e_vec, _ = _ace_constraints(
      meta_game, [0.0] * num_players, remove_null=True,
      zero_tolerance=1e-8)
  x, _ = _linear(meta_game, a_mat, e_vec, cost=cost)
  dist = np.reshape(x, meta_game.shape[1:])
  return dist, dict()


@_eliminate_dominated_decorator
def _rvce(meta_game, per_player_repeats, ignore_repeats=False):
  """Random vertex CE."""
  del ignore_repeats
  num_players = len(per_player_repeats)
  cost = np.ravel(np.random.normal(size=meta_game.shape[1:]))
  a_mat, e_vec, _ = _ace_constraints(
      meta_game, [0.0] * num_players, remove_null=True,
      zero_tolerance=1e-8)
  x, _ = _linear(meta_game, a_mat, e_vec, cost=cost)
  dist = np.reshape(x, meta_game.shape[1:])
  return dist, dict()


# Meta-solvers - CCEs.
def _mgcce(meta_game, per_player_repeats, ignore_repeats=False):
  """Maximum Gini CCE."""
  a_mat, meta = _cce_constraints(
      meta_game, [0.0] * len(per_player_repeats), remove_null=True,
      zero_tolerance=1e-8)
  a_mats = _partition_by_player(
      a_mat, meta["p_vec"], len(per_player_repeats))
  dist, _ = _try_two_solvers(
      _qp_cce,
      meta_game, a_mats, [0.0] * len(per_player_repeats),
      action_repeats=(None if ignore_repeats else per_player_repeats))
  return dist, dict()


def _min_epsilon_mgcce(meta_game, per_player_repeats, ignore_repeats=False):
  """Min Epsilon Maximum Gini CCE."""
  a_mat, meta = _cce_constraints(
      meta_game, [0.0] * len(per_player_repeats), remove_null=True,
      zero_tolerance=1e-8)
  a_mats = _partition_by_player(
      a_mat, meta["p_vec"], len(per_player_repeats))
  dist, _ = _try_two_solvers(
      _qp_cce,
      meta_game, a_mats, [0.0] * len(per_player_repeats),
      action_repeats=(None if ignore_repeats else per_player_repeats),
      min_epsilon=True)
  return dist, dict()


def _approx_mgcce(meta_game, per_player_repeats, ignore_repeats=False,
                  epsilon=0.01):
  """Maximum Gini CCE."""
  a_mat, meta = _cce_constraints(
      meta_game, [0.0] * len(per_player_repeats), remove_null=True,
      zero_tolerance=1e-8)
  max_ab = 0.0
  if a_mat.size:
    max_ab = np.max(a_mat.mean(axis=1))
  a_mat, meta = _cce_constraints(
      meta_game, [epsilon * max_ab] * len(per_player_repeats), remove_null=True,
      zero_tolerance=1e-8)
  a_mats = _partition_by_player(
      a_mat, meta["p_vec"], len(per_player_repeats))
  dist, _ = _try_two_solvers(
      _qp_cce,
      meta_game, a_mats, [0.0] * len(per_player_repeats),
      action_repeats=(None if ignore_repeats else per_player_repeats))
  return dist, dict()


def _rmwcce(meta_game, per_player_repeats, ignore_repeats=False):
  """Random maximum welfare CCE."""
  del ignore_repeats
  num_players = len(per_player_repeats)
  cost = np.ravel(np.sum(meta_game, axis=0))
  cost += np.ravel(np.random.normal(size=meta_game.shape[1:])) * 1e-6
  a_mat, _ = _cce_constraints(
      meta_game, [0.0] * num_players, remove_null=True,
      zero_tolerance=1e-8)
  e_vec = np.zeros([a_mat.shape[0]])
  x, _ = _linear(meta_game, a_mat, e_vec, cost=cost)
  dist = np.reshape(x, meta_game.shape[1:])
  return dist, dict()


def _mwcce(meta_game, per_player_repeats, ignore_repeats=False):
  """Maximum welfare CCE."""
  del ignore_repeats
  num_players = len(per_player_repeats)
  cost = np.ravel(np.sum(meta_game, axis=0))
  a_mat, _ = _cce_constraints(
      meta_game, [0.0] * num_players, remove_null=True,
      zero_tolerance=1e-8)
  e_vec = np.zeros([a_mat.shape[0]])
  x, _ = _linear(meta_game, a_mat, e_vec, cost=cost)
  dist = np.reshape(x, meta_game.shape[1:])
  return dist, dict()


def _rvcce(meta_game, per_player_repeats, ignore_repeats=False):
  """Random vertex CCE."""
  del ignore_repeats
  num_players = len(per_player_repeats)
  cost = np.ravel(np.random.normal(size=meta_game.shape[1:]))
  a_mat, _ = _cce_constraints(
      meta_game, [0.0] * num_players, remove_null=True,
      zero_tolerance=1e-8)
  e_vec = np.zeros([a_mat.shape[0]])
  x, _ = _linear(meta_game, a_mat, e_vec, cost=cost)
  dist = np.reshape(x, meta_game.shape[1:])
  return dist, dict()


# Flags to functions.
_FLAG_TO_FUNC = dict(
    uni=_uni,
    undominated_uni=_undominated_uni,
    rj=_rj,
    undominated_rj=_undominated_rj,
    rd=_rd,
    undominated_rd=_undominated_rd,
    prd=_prd,
    alpharank=_alpharank,
    mgce=_mgce,
    min_epsilon_mgce=_min_epsilon_mgce,
    approx_mgce=_approx_mgce,
    rmwce=_rmwce,
    mwce=_mwce,
    rvce=_rvce,
    mgcce=_mgcce,
    min_epsilon_mgcce=_min_epsilon_mgcce,
    approx_mgcce=_approx_mgcce,
    rmwcce=_rmwcce,
    mwcce=_mwcce,
    rvcce=_rvcce,
)


## PSRO Functions.


def intilize_policy(game, player, policy_init):
  """Returns initial policy."""
  if policy_init == "uniform":
    new_policy = policy.TabularPolicy(game, players=(player,))

  elif policy_init == "random_deterministic":
    new_policy = policy.TabularPolicy(game, players=(player,))
    for i in range(new_policy.action_probability_array.shape[0]):
      new_policy.action_probability_array[i] = np.random.multinomial(
          1, new_policy.action_probability_array[i]).astype(np.float64)

  else:
    raise ValueError(
        "policy_init must be a valid initialization strategy: %s. "
        "Received: %s" % (INIT_POLICIES, policy_init))

  return new_policy


def add_new_policies(
    per_player_new_policies,
    per_player_gaps,
    per_player_repeats,
    per_player_policies,
    joint_policies,
    joint_returns,
    game,
    br_selection):
  """Adds novel policies from new policies."""
  num_players = len(per_player_new_policies)
  per_player_num_novel_policies = [0 for _ in range(num_players)]

  # Update policies and policy counts.
  for player in range(num_players):
    new_policies = per_player_new_policies[player]
    new_gaps = per_player_gaps[player]

    repeat_policies = []
    repeat_gaps = []
    repeat_ids = []
    novel_policies = []
    novel_gaps = []
    for new_policy, new_gap in zip(new_policies, new_gaps):
      for policy_id, policy_ in enumerate(per_player_policies[player]):
        if np.all(  # New policy is not novel.
            new_policy.action_probability_array ==
            policy_.action_probability_array):  # pytype: disable=attribute-error  # py39-upgrade
          logging.debug("Player %d's new policy is not novel.", player)
          repeat_policies.append(new_policy)
          repeat_gaps.append(new_gap)
          repeat_ids.append(policy_id)
          break
      else:  # New policy is novel.
        logging.debug("Player %d's new policy is novel.", player)
        novel_policies.append(new_policy)
        novel_gaps.append(new_gap)

    add_novel_policies = []
    add_repeat_ids = []
    if (novel_policies or repeat_policies):
      if br_selection == "all":
        add_novel_policies.extend(novel_policies)
        add_repeat_ids.extend(repeat_ids)
      elif br_selection == "all_novel":
        add_novel_policies.extend(novel_policies)
      elif br_selection == "random":
        index = np.random.randint(0, len(repeat_policies) + len(novel_policies))
        if index < len(novel_policies):
          add_novel_policies.append(novel_policies[index])
        else:
          add_repeat_ids.append(repeat_ids[index - len(novel_policies)])
      elif br_selection == "random_novel":
        if novel_policies:
          index = np.random.randint(0, len(novel_policies))
          add_novel_policies.append(novel_policies[index])
        else:  # Fall back on random.
          index = np.random.randint(0, len(repeat_policies))
          add_repeat_ids.append(repeat_ids[index])
      elif br_selection == "largest_gap":
        if novel_policies:
          index = np.argmax(novel_gaps)
          if novel_gaps[index] == 0.0:  # Fall back to random when zero.
            index = np.random.randint(0, len(novel_policies))
          add_novel_policies.append(novel_policies[index])
        else:  # Fall back on random.
          index = np.random.randint(0, len(repeat_policies))
          add_repeat_ids.append(repeat_ids[index])
      else:
        raise ValueError("Unrecognized br_selection method: %s"
                         % br_selection)

    for add_repeat_id in add_repeat_ids:
      per_player_repeats[player][add_repeat_id] += 1

    for add_novel_policy in add_novel_policies:
      per_player_policies[player].append(add_novel_policy)  # Add new policy.
      per_player_repeats[player].append(1)  # Add new count.
      per_player_num_novel_policies[player] += 1

  # Add new joint policies.
  for pids in itertools.product(*[
      range(len(policies)) for policies in per_player_policies]):
    if pids in joint_policies:
      continue
    logging.debug("Evaluating novel joint policy: %s.", pids)
    policies = [
        policies[pid] for pid, policies in zip(pids, per_player_policies)]
    python_tabular_policy = policy.merge_tabular_policies(
        policies, game)
    pyspiel_tabular_policy = policy.python_policy_to_pyspiel_policy(
        python_tabular_policy)
    joint_policies[pids] = pyspiel_tabular_policy
    joint_returns[pids] = [
        0.0 if abs(er) < RETURN_TOL else er
        for er in pyspiel.expected_returns(
            game.new_initial_state(), pyspiel_tabular_policy, -1, True)]

  return per_player_num_novel_policies


def add_meta_game(
    meta_games,
    per_player_policies,
    joint_returns):
  """Returns a meta-game tensor."""
  per_player_num_policies = [
      len(policies) for policies in per_player_policies]
  shape = [len(per_player_num_policies)] + per_player_num_policies
  meta_game = np.zeros(shape)
  for pids in itertools.product(*[
      range(np_) for np_ in per_player_num_policies]):
    meta_game[(slice(None),) + pids] = joint_returns[pids]
  meta_games.append(meta_game)
  return meta_games


def add_meta_dist(
    meta_dists, meta_values, meta_solver, meta_game, per_player_repeats,
    ignore_repeats):
  """Returns meta_dist."""
  num_players = meta_game.shape[0]
  meta_solver_func = _FLAG_TO_FUNC[meta_solver]
  meta_dist, _ = meta_solver_func(
      meta_game, per_player_repeats, ignore_repeats=ignore_repeats)
  # Clean dist.
  meta_dist = meta_dist.astype(np.float64)
  meta_dist[meta_dist < DIST_TOL] = 0.0
  meta_dist[meta_dist > 1.0] = 1.0
  meta_dist /= np.sum(meta_dist)
  meta_dist[meta_dist > 1.0] = 1.0
  meta_dists.append(meta_dist)
  meta_value = np.sum(
      meta_dist * meta_game, axis=tuple(range(1, num_players + 1)))
  meta_values.append(meta_value)
  return meta_dist


def find_best_response(
    game,
    meta_dist,
    meta_game,
    iteration,
    joint_policies,
    target_equilibrium,
    update_players_strategy,
    action_value_tolerance,
):
  """Returns new best response policies."""
  num_players = meta_game.shape[0]
  per_player_num_policies = meta_dist.shape[:]

  # Player update strategy.
  if update_players_strategy == "all":
    players = list(range(num_players))
  elif update_players_strategy == "cycle":
    players = [iteration % num_players]
  elif update_players_strategy == "random":
    players = [np.random.randint(0, num_players)]
  else:
    raise ValueError(
        "update_players_strategy must be a valid player update strategy: "
        "%s. Received: %s" % (UPDATE_PLAYERS_STRATEGY, update_players_strategy))

  # Find best response.
  per_player_new_policies = []
  per_player_deviation_incentives = []

  if target_equilibrium == "cce":
    for player in range(num_players):
      if player in players:
        joint_policy_ids = itertools.product(*[
            (np_-1,) if p_ == player else range(np_) for p_, np_
            in enumerate(per_player_num_policies)])
        joint_policies_slice = [
            joint_policies[jpid] for jpid in joint_policy_ids]
        meta_dist_slice = np.sum(meta_dist, axis=player)
        meta_dist_slice[meta_dist_slice < DIST_TOL] = 0.0
        meta_dist_slice[meta_dist_slice > 1.0] = 1.0
        meta_dist_slice /= np.sum(meta_dist_slice)
        meta_dist_slice = meta_dist_slice.flat

        mu = [(p, mp) for mp, p in zip(joint_policies_slice, meta_dist_slice)
              if p > 0]
        info = pyspiel.cce_dist(
            game,
            mu,
            player,
            prob_cut_threshold=0.0,
            action_value_tolerance=action_value_tolerance)

        new_policy = policy.pyspiel_policy_to_python_policy(
            game, info.best_response_policies[0], players=(player,))
        on_policy_value = np.sum(meta_game[player] * meta_dist)
        deviation_incentive = max(
            info.best_response_values[0] - on_policy_value, 0)
        if deviation_incentive < GAP_TOL:
          deviation_incentive = 0.0

        per_player_new_policies.append([new_policy])
        per_player_deviation_incentives.append([deviation_incentive])
      else:
        per_player_new_policies.append([])
        per_player_deviation_incentives.append([])

  elif target_equilibrium == "ce":
    for player in range(num_players):
      if player in players:
        per_player_new_policies.append([])
        per_player_deviation_incentives.append([])

        for pid in range(per_player_num_policies[player]):
          joint_policy_ids = itertools.product(*[
              (pid,) if p_ == player else range(np_) for p_, np_
              in enumerate(per_player_num_policies)])
          joint_policies_slice = [
              joint_policies[jpid] for jpid in joint_policy_ids]
          inds = tuple((pid,) if player == p_ else slice(None)
                       for p_ in range(num_players))
          meta_dist_slice = np.ravel(meta_dist[inds]).copy()
          meta_dist_slice[meta_dist_slice < DIST_TOL] = 0.0
          meta_dist_slice[meta_dist_slice > 1.0] = 1.0
          meta_dist_slice_sum = np.sum(meta_dist_slice)

          if meta_dist_slice_sum > 0.0:
            meta_dist_slice /= meta_dist_slice_sum
            mu = [(p, mp) for mp, p in
                  zip(joint_policies_slice, meta_dist_slice)
                  if p > 0]
            info = pyspiel.cce_dist(
                game,
                mu,
                player,
                prob_cut_threshold=0.0,
                action_value_tolerance=action_value_tolerance)

            new_policy = policy.pyspiel_policy_to_python_policy(
                game, info.best_response_policies[0], players=(player,))
            on_policy_value = np.sum(
                np.ravel(meta_game[player][inds]) * meta_dist_slice)
            deviation_incentive = max(
                info.best_response_values[0] - on_policy_value, 0)
            if deviation_incentive < GAP_TOL:
              deviation_incentive = 0.0

            per_player_new_policies[-1].append(new_policy)
            per_player_deviation_incentives[-1].append(
                meta_dist_slice_sum * deviation_incentive)

      else:
        per_player_new_policies.append([])
        per_player_deviation_incentives.append([])

  else:
    raise ValueError(
        "target_equilibrium must be a valid best response strategy: %s. "
        "Received: %s" % (BRS, target_equilibrium))

  return per_player_new_policies, per_player_deviation_incentives


## Main Loop.


def initialize(game, train_meta_solver, eval_meta_solver, policy_init,
               ignore_repeats, br_selection):
  """Return initialized data structures."""
  num_players = game.num_players()

  # Initialize.
  iteration = 0
  per_player_repeats = [[] for _ in range(num_players)]
  per_player_policies = [[] for _ in range(num_players)]
  joint_policies = {}  # Eg. (1, 0): Joint policy.
  joint_returns = {}
  meta_games = []
  train_meta_dists = []
  eval_meta_dists = []
  train_meta_values = []
  eval_meta_values = []
  train_meta_gaps = []
  eval_meta_gaps = []

  # Initialize policies.
  per_player_new_policies = [
      [intilize_policy(game, player, policy_init)]
      for player in range(num_players)]
  per_player_gaps_train = [[1.0] for player in range(num_players)]
  per_player_num_novel_policies = add_new_policies(
      per_player_new_policies, per_player_gaps_train, per_player_repeats,
      per_player_policies, joint_policies, joint_returns, game, br_selection)
  del per_player_num_novel_policies
  add_meta_game(
      meta_games,
      per_player_policies,
      joint_returns)
  add_meta_dist(
      train_meta_dists, train_meta_values, train_meta_solver,
      meta_games[-1], per_player_repeats, ignore_repeats)
  add_meta_dist(
      eval_meta_dists, eval_meta_values, eval_meta_solver,
      meta_games[-1], per_player_repeats, ignore_repeats)

  return (
      iteration,
      per_player_repeats,
      per_player_policies,
      joint_policies,
      joint_returns,
      meta_games,
      train_meta_dists,
      eval_meta_dists,
      train_meta_values,
      eval_meta_values,
      train_meta_gaps,
      eval_meta_gaps)


def initialize_callback_(
    iteration,
    per_player_repeats,
    per_player_policies,
    joint_policies,
    joint_returns,
    meta_games,
    train_meta_dists,
    eval_meta_dists,
    train_meta_values,
    eval_meta_values,
    train_meta_gaps,
    eval_meta_gaps,
    game):
  """Callback which allows initializing from checkpoint."""
  del game
  checkpoint = None
  return (
      iteration,
      per_player_repeats,
      per_player_policies,
      joint_policies,
      joint_returns,
      meta_games,
      train_meta_dists,
      eval_meta_dists,
      train_meta_values,
      eval_meta_values,
      train_meta_gaps,
      eval_meta_gaps,
      checkpoint)


def callback_(
    iteration,
    per_player_repeats,
    per_player_policies,
    joint_policies,
    joint_returns,
    meta_games,
    train_meta_dists,
    eval_meta_dists,
    train_meta_values,
    eval_meta_values,
    train_meta_gaps,
    eval_meta_gaps,
    kwargs,
    checkpoint):
  """Callback for updating checkpoint."""
  del iteration, per_player_repeats, per_player_policies, joint_policies
  del joint_returns, meta_games, train_meta_dists, eval_meta_dists
  del train_meta_values, eval_meta_values, train_meta_gaps, eval_meta_gaps
  del kwargs
  return checkpoint


def run_loop(game,
             game_name,
             seed=0,
             iterations=40,
             policy_init="uniform",
             update_players_strategy="all",
             target_equilibrium="cce",
             br_selection="largest_gap",
             train_meta_solver="mgcce",
             eval_meta_solver="mwcce",
             ignore_repeats=False,
             initialize_callback=None,
             action_value_tolerance=-1.0,
             callback=None):
  """Runs JPSRO."""
  if initialize_callback is None:
    initialize_callback = initialize_callback_
  if callback is None:
    callback = callback_
  kwargs = dict(
      game=game,
      game_name=game_name,
      seed=seed,
      iterations=iterations,
      policy_init=policy_init,
      update_players_strategy=update_players_strategy,
      target_equilibrium=target_equilibrium,
      br_selection=br_selection,
      train_meta_solver=train_meta_solver,
      eval_meta_solver=eval_meta_solver,
      ignore_repeats=ignore_repeats,
  )

  # Set seed.
  np.random.seed(seed)

  # Some statistics.
  num_players = game.num_players()  # Look in the game.

  # Initialize.
  values = initialize(game, train_meta_solver, eval_meta_solver, policy_init,
                      ignore_repeats, br_selection)

  # Initialize Callback.
  (iteration,
   per_player_repeats,
   per_player_policies,
   joint_policies,
   joint_returns,
   meta_games,
   train_meta_dists,
   eval_meta_dists,
   train_meta_values,
   eval_meta_values,
   train_meta_gaps,
   eval_meta_gaps,
   checkpoint) = initialize_callback(*values, game)

  # Run JPSRO.
  while iteration <= iterations:
    logging.debug("Beginning JPSRO iteration %03d", iteration)
    per_player_new_policies, per_player_gaps_train = find_best_response(
        game,
        train_meta_dists[-1],
        meta_games[-1],
        iteration,
        joint_policies,
        target_equilibrium,
        update_players_strategy,
        action_value_tolerance,
    )
    train_meta_gaps.append([sum(gaps) for gaps in per_player_gaps_train])
    _, per_player_gaps_eval = find_best_response(
        game,
        eval_meta_dists[-1],
        meta_games[-1],
        iteration,
        joint_policies,
        target_equilibrium,
        update_players_strategy,
        action_value_tolerance,
    )
    eval_meta_gaps.append([sum(gaps) for gaps in per_player_gaps_eval])
    per_player_num_novel_policies = add_new_policies(
        per_player_new_policies, per_player_gaps_train, per_player_repeats,
        per_player_policies, joint_policies, joint_returns, game, br_selection)
    del per_player_num_novel_policies
    add_meta_game(
        meta_games,
        per_player_policies,
        joint_returns)
    add_meta_dist(
        train_meta_dists, train_meta_values, train_meta_solver,
        meta_games[-1], per_player_repeats, ignore_repeats)
    add_meta_dist(
        eval_meta_dists, eval_meta_values, eval_meta_solver,
        meta_games[-1], per_player_repeats, ignore_repeats)

    # Stats.
    per_player_num_policies = train_meta_dists[-1].shape[:]
    log_string = LOG_STRING.format(
        iteration=iteration,
        game=game_name,
        player=("{: 12d}" * num_players).format(*list(range(num_players))),
        brs="",
        num_policies=("{: 12d}" * num_players).format(*[
            sum(ppr) for ppr in per_player_repeats]),
        unique=("{: 12d}" * num_players).format(*per_player_num_policies),
        train_meta_solver=train_meta_solver,
        train_value=("{: 12g}" * num_players).format(*train_meta_values[-1]),
        train_gap=("{: 12g}" * num_players).format(*train_meta_gaps[-1]),
        eval_meta_solver=eval_meta_solver,
        eval_value=("{: 12g}" * num_players).format(*eval_meta_values[-1]),
        eval_gap=("{: 12g}" * num_players).format(*eval_meta_gaps[-1]),
    )
    logging.info(log_string)

    # Increment.
    iteration += 1

    # Callback.
    checkpoint = callback(
        iteration,
        per_player_repeats,
        per_player_policies,
        joint_policies,
        joint_returns,
        meta_games,
        train_meta_dists,
        eval_meta_dists,
        train_meta_values,
        eval_meta_values,
        train_meta_gaps,
        eval_meta_gaps,
        kwargs,
        checkpoint)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Mean Field Crowd Modelling, implemented in Python.

This is a demonstration of implementing a mean field game in Python.

Fictitious play for mean field games: Continuous time analysis and applications,
Perrin & al. 2019 (https://arxiv.org/abs/2007.03458). This game corresponds
to the game in section 4.2.
"""

from typing import Any, List, Mapping
import numpy as np

from open_spiel.python.observation import IIGObserverForPublicInfoGame
import pyspiel

_NUM_PLAYERS = 1
_SIZE = 10
_HORIZON = 10
_NUM_ACTIONS = 3
_NUM_CHANCE = 3
_EPSILON = 10**(-25)
_DEFAULT_PARAMS = {"size": _SIZE, "horizon": _HORIZON}
_GAME_TYPE = pyspiel.GameType(
    short_name="python_mfg_crowd_modelling",
    long_name="Python Mean Field Crowd Modelling",
    dynamics=pyspiel.GameType.Dynamics.MEAN_FIELD,
    chance_mode=pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC,
    information=pyspiel.GameType.Information.PERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.GENERAL_SUM,
    reward_model=pyspiel.GameType.RewardModel.REWARDS,
    max_num_players=_NUM_PLAYERS,
    min_num_players=_NUM_PLAYERS,
    provides_information_state_string=True,
    provides_information_state_tensor=False,
    provides_observation_string=True,
    provides_observation_tensor=True,
    parameter_specification=_DEFAULT_PARAMS)


class MFGCrowdModellingGame(pyspiel.Game):
  """A Mean Field Crowd Modelling game.


    A game starts by an initial chance node that select the initial state
    of the MFG.
    Then the game sequentially alternates between:
      - An action selection node (Where the player Id >= 0)
      - A chance node (the player id is pyspiel.PlayerId.CHANCE)
      - A Mean Field node (the player id is pyspiel.PlayerId.MEAN_FIELD)
  """

  # pylint:disable=dangerous-default-value
  def __init__(self, params: Mapping[str, Any] = _DEFAULT_PARAMS):
    game_info = pyspiel.GameInfo(
        num_distinct_actions=_NUM_ACTIONS,
        max_chance_outcomes=max(params["size"], _NUM_CHANCE),
        num_players=_NUM_PLAYERS,
        min_utility=-np.inf,
        max_utility=+np.inf,
        utility_sum=None,
        max_game_length=params["horizon"])
    super().__init__(_GAME_TYPE, game_info, params)
    self.size = params["size"]
    self.horizon = params["horizon"]

  def new_initial_state(self):
    """Returns a state corresponding to the start of a game."""
    return MFGCrowdModellingState(self)

  def make_py_observer(self, iig_obs_type=None, params=None):
    """Returns an object used for observing game state."""
    if ((iig_obs_type is None) or
        (iig_obs_type.public_info and not iig_obs_type.perfect_recall)):
      return Observer(params, self)
    return IIGObserverForPublicInfoGame(iig_obs_type, params)

  def max_chance_nodes_in_history(self):
    """Maximun chance nodes in game history."""
    return self.horizon + 1


class MFGCrowdModellingState(pyspiel.State):
  """A Mean Field Crowd Modelling state."""

  # Maps legal actions to the corresponding move along the 1-D axis of the game.
  _ACTION_TO_MOVE = {0: -1, 1: 0, 2: 1}
  # Action that corresponds to no displacement.
  _NEUTRAL_ACTION = 1

  def __init__(self, game):
    """Constructor; should only be called by Game.new_initial_state."""
    super().__init__(game)
    self._is_chance_init = True  # is true for the first state of the game.
    self._player_id = pyspiel.PlayerId.CHANCE
    self._x = None
    self._t = 0
    # We initialize last_action to the neutral action. This makes sure
    # that the first reward does not include any displacement penalty.
    self._last_action = self._NEUTRAL_ACTION
    self.size = game.size
    self.horizon = game.horizon
    self.return_value = 0.0

    # Represents the current probability distribution over game states.
    # Initialized with a uniform distribution.
    self._distribution = [1. / self.size for i in range(self.size)]

  @property
  def x(self):
    return self._x

  @property
  def t(self):
    return self._t

  def state_to_str(self, x, t, player_id=pyspiel.PlayerId.DEFAULT_PLAYER_ID):
    """A string that uniquely identify a triplet x, t, player_id."""
    if self._is_chance_init:
      return "initial"
    if player_id == pyspiel.PlayerId.DEFAULT_PLAYER_ID:
      return str((x, t))
    if player_id == pyspiel.PlayerId.MEAN_FIELD:
      return str((x, t)) + "_a"
    if player_id == pyspiel.PlayerId.CHANCE:
      return str((x, t)) + "_a_mu"
    raise ValueError(
        "player_id is not mean field, chance or default player id.")

  # OpenSpiel (PySpiel) API functions are below. This is the standard set that
  # should be implemented by every perfect-information sequential-move game.

  def _legal_actions(self, player):
    """Returns a list of legal actions for player and MFG nodes."""
    if player == pyspiel.PlayerId.MEAN_FIELD:
      return []
    if (player == pyspiel.PlayerId.DEFAULT_PLAYER_ID
        and player == self.current_player()):
      return [0, 1, 2]
    raise ValueError(f"Unexpected player {player}. "
                     "Expected a mean field or current player 0.")

  def chance_outcomes(self):
    """Returns the possible chance outcomes and their probabilities."""
    if self._is_chance_init:
      return list(enumerate(self._distribution))
    return [(0, 1. / 3.), (1, 1. / 3.), (2, 1. / 3.)]

  def _apply_action(self, action):
    """Applies the specified action to the state."""
    if self._player_id == pyspiel.PlayerId.MEAN_FIELD:
      raise ValueError(
          "_apply_action should not be called at a MEAN_FIELD state.")
    self.return_value += self._rewards()
    if self._is_chance_init:
      # Here the action is between 0 and self.size - 1
      if action < 0 or action >= self.size:
        raise ValueError(
            "The action is between 0 and self.size - 1 at an init chance node")
      self._x = action
      self._is_chance_init = False
      self._player_id = pyspiel.PlayerId.DEFAULT_PLAYER_ID
    elif self._player_id == pyspiel.PlayerId.CHANCE:
      # Here the action is between 0 and 2
      if action < 0 or action > 2:
        raise ValueError(
            "The action is between 0 and 2 at any chance node")
      self._x = (self.x + self._ACTION_TO_MOVE[action]) % self.size
      self._t += 1
      self._player_id = pyspiel.PlayerId.MEAN_FIELD
    elif self._player_id == pyspiel.PlayerId.DEFAULT_PLAYER_ID:
      # Here the action is between 0 and 2
      if action < 0 or action > 2:
        raise ValueError(
            "The action is between 0 and 2 at any chance node")
      self._x = (self.x + self._ACTION_TO_MOVE[action]) % self.size
      self._last_action = action
      self._player_id = pyspiel.PlayerId.CHANCE

  def _action_to_string(self, player, action):
    """Action -> string."""
    del player
    if self.is_chance_node() and self._is_chance_init:
      return f"init_state={action}"
    return str(self._ACTION_TO_MOVE[action])

  def distribution_support(self):
    """return a list of state string."""
    return [
        self.state_to_str(
            i, self.t, player_id=pyspiel.PlayerId.MEAN_FIELD)
        for i in range(self.size)
    ]

  def update_distribution(self, distribution):
    """This function is central and specific to the logic of the MFG.

    Args:
      distribution: a distribution to register.

      - function should be called when the node is in MEAN_FIELD state.
      - distribution are probabilities that correspond to each game state
      given by distribution_support.

    """
    if self._player_id != pyspiel.PlayerId.MEAN_FIELD:
      raise ValueError(
          "update_distribution should only be called at a MEAN_FIELD state.")
    self._distribution = distribution.copy()
    self._player_id = pyspiel.PlayerId.DEFAULT_PLAYER_ID

  def is_terminal(self):
    """Returns True if the game is over."""
    return self.t >= self.horizon

  def current_player(self):
    """Returns id of the next player to move, or TERMINAL if game is over."""
    if self.is_terminal():
      return pyspiel.PlayerId.TERMINAL
    return self._player_id

  def _rewards(self):
    """Reward for the player for this state."""
    if self._player_id == pyspiel.PlayerId.DEFAULT_PLAYER_ID:
      r_x = 1 - (1.0 * np.abs(self.x - self.size // 2)) / (self.size // 2)
      r_a = -(1.0 * np.abs(self._ACTION_TO_MOVE[self._last_action])) / self.size
      r_mu = - np.log(self._distribution[self.x] + _EPSILON)
      return r_x + r_a + r_mu
    return 0.0

  def rewards(self) -> List[float]:
    """Rewards for all players."""
    # For now, only single-population (single-player) mean field games
    # are supported.
    return [self._rewards()]

  def _returns(self):
    """Returns is the sum of all payoffs collected so far."""
    return self.return_value + self._rewards()

  def returns(self) -> List[float]:
    """Returns for all players."""
    # For now, only single-population (single-player) mean field games
    # are supported.
    return [self._returns()]

  def __str__(self):
    """A string that uniquely identify the current state."""
    return self.state_to_str(self.x, self.t, player_id=self._player_id)


class Observer:
  """Observer, conforming to the PyObserver interface (see observation.py)."""

  def __init__(self, params, game):
    """Initializes an empty observation tensor."""
    del params

    self.size = game.size
    self.horizon = game.horizon
    # +1 to allow t == horizon.
    self.tensor = np.zeros(self.size + self.horizon + 1, np.float32)
    self.dict = {"x": self.tensor[:self.size], "t": self.tensor[self.size:]}

  def set_from(self, state: MFGCrowdModellingState, player: int):
    """Updates `tensor` and `dict` to reflect `state` from PoV of `player`."""
    del player
    # We update the observation via the shaped tensor since indexing is more
    # convenient than with the 1-D tensor. Both are views onto the same memory.
    self.tensor.fill(0)
    # state.x is None for the initial (blank) state, don't set any
    # position bit in that case.
    if state.x is not None:
      if not 0 <= state.x < self.size:
        raise ValueError(
            f"Expected {state} x position to be in [0, {self.size})")
      self.dict["x"][state.x] = 1
    if not 0 <= state.t <= self.horizon:
      raise ValueError(f"Expected {state} time to be in [0, {self.horizon}]")
    self.dict["t"][state.t] = 1

  def string_from(self, state, player):
    """Observation of `state` from the PoV of `player`, as a string."""
    del player
    return str(state)


# Register the game with the OpenSpiel library

pyspiel.register_game(_GAME_TYPE, MFGCrowdModellingGame)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for open_spiel.python.pybind11.pyspiel."""

import os
from absl.testing import absltest

from open_spiel.python import games  # pylint: disable=unused-import
from open_spiel.python import policy
from open_spiel.python.mfg import games as mfgs  # pylint: disable=unused-import
import pyspiel

# Specify game names in alphabetical order, to make the test easier to read.
EXPECTED_GAMES = frozenset([
    "2048",
    "add_noise",
    "amazons",
    "backgammon",
    "bargaining",
    "battleship",
    "blackjack",
    "blotto",
    "breakthrough",
    "bridge",
    "bridge_uncontested_bidding",
    "catch",
    "checkers",
    "chess",
    "cliff_walking",
    "clobber",
    "coin_game",
    "colored_trails",
    "connect_four",
    "coop_box_pushing",
    "coop_to_1p",
    "coordinated_mp",
    "crazy_eights",
    "cursor_go",
    "dark_chess",
    "dark_hex",
    "dark_hex_ir",
    "deep_sea",
    "dou_dizhu",
    "efg_game",
    "euchre",
    "first_sealed_auction",
    "gin_rummy",
    "go",
    "goofspiel",
    "havannah",
    "hex",
    "hearts",
    "kriegspiel",
    "kuhn_poker",
    "laser_tag",
    "lewis_signaling",
    "leduc_poker",
    "liars_dice",
    "liars_dice_ir",
    "maedn",
    "mancala",
    "markov_soccer",
    "matching_pennies_3p",
    "matrix_bos",
    "matrix_brps",
    "matrix_cd",
    "matrix_coordination",
    "matrix_mp",
    "matrix_pd",
    "matrix_rps",
    "matrix_rpsw",
    "matrix_sh",
    "matrix_shapleys_game",
    "mean_field_lin_quad",
    "mfg_crowd_modelling",
    "mfg_crowd_modelling_2d",
    "mfg_dynamic_routing",
    "mfg_garnet",
    "misere",
    "morpion_solitaire",
    "negotiation",
    "nfg_game",
    "nim",
    "nine_mens_morris",
    "normal_form_extensive_game",
    "oh_hell",
    "oshi_zumo",
    "othello",
    "oware",
    "pentago",
    "pathfinding",
    "phantom_go",
    "phantom_ttt",
    "phantom_ttt_ir",
    "pig",
    "python_block_dominoes",
    "python_dynamic_routing",
    "python_iterated_prisoners_dilemma",
    "python_mfg_crowd_modelling",
    "python_mfg_dynamic_routing",
    "python_mfg_predator_prey",
    "python_kuhn_poker",
    "python_tic_tac_toe",
    "python_liars_poker",
    "quoridor",
    "repeated_game",
    "rbc",
    "restricted_nash_response",
    "sheriff",
    "skat",
    "start_at",
    "solitaire",
    "stones_and_gems",
    "tarok",
    "tic_tac_toe",
    "tiny_bridge_2p",
    "tiny_bridge_4p",
    "tiny_hanabi",
    "trade_comm",
    "turn_based_simultaneous_game",
    "ultimate_tic_tac_toe",
    "y",
    "zerosum",
])


class PyspielTest(absltest.TestCase):

  def test_registered_names(self):
    game_names = pyspiel.registered_names()

    expected = list(EXPECTED_GAMES)
    if (os.environ.get("OPEN_SPIEL_BUILD_WITH_HANABI", "OFF") == "ON" and
        "hanabi" not in expected):
      expected.append("hanabi")
    if (os.environ.get("OPEN_SPIEL_BUILD_WITH_ACPC", "OFF") == "ON" and
        "universal_poker" not in expected):
      expected.append("universal_poker")
    expected = sorted(expected)
    self.assertCountEqual(game_names, expected)

  def teste_default_loadable(self):
    # Games which cannmot be loaded with default parameters will be skipped by
    # several standard tests. We make a list of such games here in order to make
    # implementors think twice about making new games non-default-loadable
    non_default_loadable = [
        game.short_name
        for game in pyspiel.registered_games()
        if not game.default_loadable
    ]
    expected = [
        # Being non-default-loadable prevents various automated tests.
        # Only add games here if there is no sensible default for a parameter.
        "add_noise",
        "efg_game",
        "nfg_game",
        "misere",
        "turn_based_simultaneous_game",
        "normal_form_extensive_game",
        "repeated_game",
        "restricted_nash_response",
        "start_at",
        "zerosum",
    ]
    self.assertCountEqual(non_default_loadable, expected)

  def test_registered_game_attributes(self):
    game_list = {game.short_name: game for game in pyspiel.registered_games()}
    self.assertEqual(game_list["kuhn_poker"].dynamics,
                     pyspiel.GameType.Dynamics.SEQUENTIAL)
    self.assertEqual(game_list["kuhn_poker"].chance_mode,
                     pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC)
    self.assertEqual(game_list["kuhn_poker"].information,
                     pyspiel.GameType.Information.IMPERFECT_INFORMATION)
    self.assertEqual(game_list["kuhn_poker"].utility,
                     pyspiel.GameType.Utility.ZERO_SUM)
    self.assertEqual(game_list["kuhn_poker"].min_num_players, 2)

  def test_create_game(self):
    game = pyspiel.load_game("kuhn_poker")
    game_info = game.get_type()
    self.assertEqual(game_info.information,
                     pyspiel.GameType.Information.IMPERFECT_INFORMATION)
    self.assertEqual(game.num_players(), 2)

  def test_play_kuhn_poker(self):
    game = pyspiel.load_game("kuhn_poker")
    state = game.new_initial_state()
    self.assertEqual(state.is_chance_node(), True)
    self.assertEqual(state.chance_outcomes(), [(0, 1 / 3), (1, 1 / 3),
                                               (2, 1 / 3)])
    state.apply_action(1)
    self.assertEqual(state.is_chance_node(), True)
    self.assertEqual(state.chance_outcomes(), [(0, 0.5), (2, 0.5)])
    state.apply_action(2)
    self.assertEqual(state.is_chance_node(), False)
    self.assertEqual(state.legal_actions(), [0, 1])
    sampler = pyspiel.UniformProbabilitySampler(0., 1.)
    clone = state.resample_from_infostate(1, sampler)
    self.assertEqual(
        clone.information_state_string(1), state.information_state_string(1))

  def test_othello(self):
    game = pyspiel.load_game("othello")
    state = game.new_initial_state()
    self.assertFalse(state.is_chance_node())
    self.assertFalse(state.is_terminal())
    self.assertEqual(state.legal_actions(), [19, 26, 37, 44])

  def test_tic_tac_toe(self):
    game = pyspiel.load_game("tic_tac_toe")
    state = game.new_initial_state()
    self.assertFalse(state.is_chance_node())
    self.assertFalse(state.is_terminal())
    self.assertEqual(state.legal_actions(), [0, 1, 2, 3, 4, 5, 6, 7, 8])

  def test_game_parameters_from_string_empty(self):
    self.assertEqual(pyspiel.game_parameters_from_string(""), {})

  def test_game_parameters_from_string_simple(self):
    self.assertEqual(
        pyspiel.game_parameters_from_string("foo"), {"name": "foo"})

  def test_game_parameters_from_string_with_options(self):
    self.assertEqual(
        pyspiel.game_parameters_from_string("foo(x=2,y=true)"), {
            "name": "foo",
            "x": 2,
            "y": True
        })

  def test_game_parameters_from_string_with_subgame(self):
    self.assertEqual(
        pyspiel.game_parameters_from_string(
            "foo(x=2,y=true,subgame=bar(z=False))"), {
                "name": "foo",
                "x": 2,
                "y": True,
                "subgame": {
                    "name": "bar",
                    "z": False
                }
            })

  def test_game_parameters_to_string_empty(self):
    self.assertEqual(pyspiel.game_parameters_to_string({}), "")

  def test_game_parameters_to_string_simple(self):
    self.assertEqual(
        pyspiel.game_parameters_to_string({"name": "foo"}), "foo()")

  def test_game_parameters_to_string_with_options(self):
    self.assertEqual(
        pyspiel.game_parameters_to_string({
            "name": "foo",
            "x": 2,
            "y": True
        }), "foo(x=2,y=True)")

  def test_game_parameters_to_string_with_subgame(self):
    self.assertEqual(
        pyspiel.game_parameters_to_string({
            "name": "foo",
            "x": 2,
            "y": True,
            "subgame": {
                "name": "bar",
                "z": False
            }
        }), "foo(subgame=bar(z=False),x=2,y=True)")

  def test_game_type(self):
    game_type = pyspiel.GameType(
        "matrix_mp", "Matching Pennies", pyspiel.GameType.Dynamics.SIMULTANEOUS,
        pyspiel.GameType.ChanceMode.DETERMINISTIC,
        pyspiel.GameType.Information.PERFECT_INFORMATION,
        pyspiel.GameType.Utility.ZERO_SUM,
        pyspiel.GameType.RewardModel.TERMINAL, 2, 2, True, True, False, False,
        dict())
    self.assertEqual(game_type.chance_mode,
                     pyspiel.GameType.ChanceMode.DETERMINISTIC)

  def test_error_handling(self):
    with self.assertRaisesRegex(RuntimeError,
                                "Unknown game 'invalid_game_name'"):
      unused_game = pyspiel.load_game("invalid_game_name")

  def test_can_create_cpp_tabular_policy(self):
    for game_name in ["kuhn_poker", "leduc_poker", "liars_dice"]:
      game = pyspiel.load_game(game_name)

      # We just test that we can create a tabular policy.
      policy.python_policy_to_pyspiel_policy(policy.TabularPolicy(game))

  def test_simultaneous_game_history(self):
    game = pyspiel.load_game("coop_box_pushing")
    state = game.new_initial_state()
    state.apply_action(0)
    state2 = game.new_initial_state()
    state2.apply_actions([0] * game.num_players())
    self.assertEqual(state.history(), state2.history())

  def test_record_batched_trajectories(self):
    for game_name in ["kuhn_poker", "leduc_poker", "liars_dice"]:
      game = pyspiel.load_game(game_name)
      python_policy = policy.TabularPolicy(game)
      tabular_policy = policy.python_policy_to_pyspiel_policy(python_policy)
      policies = [tabular_policy] * 2

      # We test that we can create a batch of trajectories.
      seed = 0
      batch_size = 128
      include_full_observations = False
      pyspiel.record_batched_trajectories(game, policies,
                                          python_policy.state_lookup,
                                          batch_size, include_full_observations,
                                          seed, -1)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Mean Field predator-prey game, implemented in Python.

This corresponds to the predator-prey game described in section 5.4 of
"Scaling up Mean Field Games with Online Mirror Descent"
(https://arxiv.org/abs/2103.00623)

The environment is configurable in the following high-level ways:
- Number of populations.
- Reward matrix.
- Geometry (torus, basic square).
"""

import enum
import functools
from typing import Any, List, Mapping, Optional, Tuple
import numpy as np

from open_spiel.python.observation import IIGObserverForPublicInfoGame
import pyspiel
from open_spiel.python.utils import shared_value


class Geometry(enum.IntEnum):
  SQUARE = 0
  TORUS = 1


_DEFAULT_SIZE = 5
_DEFAULT_HORIZON = 10
_NUM_ACTIONS = 5
_NUM_CHANCE = 5
_DEFAULT_REWARD_MATRIX = np.array([[0, -1, 1], [1, 0, -1], [-1, 1, 0]])
_DEFAULT_NUM_PLAYERS = 3

_DEFAULT_GEOMETRY = Geometry.SQUARE

_DEFAULT_PARAMS = {
    "size": _DEFAULT_SIZE,
    "horizon": _DEFAULT_HORIZON,
    "players": _DEFAULT_NUM_PLAYERS,
    # The reward matrix is represented as a string containing a
    # space-separated list of values.
    # Its size defines the number of populations in the mean field game.
    "reward_matrix": " ".join(str(v) for v in _DEFAULT_REWARD_MATRIX.flatten()),
    "geometry": _DEFAULT_GEOMETRY
}

_GAME_TYPE = pyspiel.GameType(
    short_name="python_mfg_predator_prey",
    long_name="Python Mean Field Predator Prey",
    dynamics=pyspiel.GameType.Dynamics.MEAN_FIELD,
    chance_mode=pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC,
    information=pyspiel.GameType.Information.PERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.GENERAL_SUM,
    reward_model=pyspiel.GameType.RewardModel.REWARDS,
    # We cannot pass math.inf here, so we pass a very high integer value.
    max_num_players=1000000000,
    min_num_players=1,
    provides_information_state_string=True,
    provides_information_state_tensor=False,
    provides_observation_string=True,
    provides_observation_tensor=True,
    parameter_specification=_DEFAULT_PARAMS)


def get_param(param_name, params):
  return params.get(param_name, _DEFAULT_PARAMS[param_name])


@functools.lru_cache(maxsize=None)
def _state_to_str(x, y, t, population, player_id):
  """A string that uniquely identify (pos, t, population, player_id)."""
  if int(player_id) >= 0:
    return f"(pop={population}, t={t}, pos=[{x} {y}])"
  if player_id == pyspiel.PlayerId.MEAN_FIELD:
    return f"(pop={population}, t={t}_a, pos=[{x} {y}])"
  if player_id == pyspiel.PlayerId.CHANCE:
    return f"(pop={population}, t={t}_a_mu, pos=[{x} {y}])"


class MFGPredatorPreyGame(pyspiel.Game):
  """Predator-prey multi-population MFG."""

  # pylint:disable=dangerous-default-value
  def __init__(self, params: Mapping[str, Any] = _DEFAULT_PARAMS):
    self.size = get_param("size", params)
    self.horizon = get_param("horizon", params)
    flat_reward_matrix = np.fromstring(
        get_param("reward_matrix", params), dtype=np.float64, sep=" ")
    num_players = get_param("players", params)
    if len(flat_reward_matrix) != num_players**2:
      raise ValueError(
          f"Reward matrix passed in flat representation does not represent a "
          f"square matrix: {flat_reward_matrix}")
    self.reward_matrix = flat_reward_matrix.reshape([num_players, num_players])
    self.geometry = get_param("geometry", params)
    num_states = self.size**2
    game_info = pyspiel.GameInfo(
        num_distinct_actions=_NUM_ACTIONS,
        max_chance_outcomes=max(num_states, _NUM_CHANCE),
        num_players=num_players,
        min_utility=-np.inf,
        max_utility=+np.inf,
        utility_sum=None,
        max_game_length=self.horizon)

    # Represents the current probability distribution over game states
    # (when grouped for each population). Initialized with a uniform
    # distribution.
    self.initial_distribution = [1. / num_states] * (num_states * num_players)
    super().__init__(_GAME_TYPE, game_info, params)

  def new_initial_state(self):
    """Returns a new population-less blank state.

    This state is provided for some internal operations that use blank
    states (e.g. cloning), but cannot be used to play the game, i.e.
    ApplyAction() will fail. Proper playable states should be
    instantiated with new_initial_state_for_population().
    """
    return MFGPredatorPreyState(self)

  def max_chance_nodes_in_history(self):
    """Maximun chance nodes in game history."""
    return self.horizon + 1

  def new_initial_state_for_population(self, population):
    """State corresponding to the start of a game for a given population."""
    return MFGPredatorPreyState(self, population)

  def make_py_observer(self, iig_obs_type=None, params=None):
    """Returns an object used for observing game state."""
    if ((iig_obs_type is None) or
        (iig_obs_type.public_info and not iig_obs_type.perfect_recall)):
      return Observer(params, self)
    return IIGObserverForPublicInfoGame(iig_obs_type, params)


def pos_to_merged(pos: np.ndarray, size: int) -> int:
  """Converts a [x, y] position into a single integer."""
  assert (pos >= 0).all(), pos
  assert (pos < size).all(), pos
  return pos[0] + pos[1] * size


def merged_to_pos(merged_pos: int, size: int) -> np.ndarray:
  """Inverse of pos_to_merged()."""
  assert 0 <= merged_pos < size * size
  return np.array([merged_pos % size, merged_pos // size])


class MFGPredatorPreyState(pyspiel.State):
  """State for the predator-prey MFG."""

  # Maps legal actions to the corresponding move on the grid of the game.
  _ACTION_TO_MOVE = {
      0: np.array([0, 0]),
      1: np.array([1, 0]),
      2: np.array([0, 1]),
      3: np.array([0, -1]),
      4: np.array([-1, 0]),
  }
  # Action that corresponds to no displacement.
  _NEUTRAL_ACTION = 0

  def __init__(self, game, population=None):
    """Constructor; should only be called by Game.new_initial_state.*.

    Args:
      game: MFGPredatorPreyGame for which a state should be created.
      population: ID of the population to create this state for. Must be in [0,
        num_players()) or None. States with population=None cannot be used to
        perform game actions.
    """
    super().__init__(game)
    # Initial state where the initial position is chosen according to
    # an initial distribution.
    self._is_position_init = True
    self._player_id = pyspiel.PlayerId.CHANCE
    # Population this state corresponds to. Can be None, in which
    # case, ApplyAction() is forbidden.
    self._population = population
    if self._population is not None:
      assert 0 <= self._population < self.num_players()
    # When set, <int>[2] numpy array representing the x, y position on the grid.
    self._pos = None  # type: Optional[np.ndarray]
    self._t = 0
    self.size = game.size
    # Number of states in the grid.
    self.num_states = self.size**2
    self.horizon = game.horizon
    self.reward_matrix = game.reward_matrix
    self.geometry = game.geometry
    self._returns = np.zeros([self.num_players()], dtype=np.float64)
    self._distribution = shared_value.SharedValue(game.initial_distribution)

  @property
  def pos(self):
    return self._pos

  @property
  def t(self):
    return self._t

  def state_to_str(self, pos, t, population, player_id=0):
    """A string that uniquely identify (pos, t, population, player_id)."""
    if self._is_position_init:
      return f"position_init_{population}"
    assert isinstance(pos, np.ndarray), f"Got type {type(pos)}"
    assert len(pos.shape) == 1, f"Got {len(pos.shape)}, expected 1 (pos={pos})."
    assert pos.shape[0] == 2, f"Got {pos.shape[0]}, expected 2 (pos={pos})."
    return _state_to_str(pos[0], pos[1], t, population, player_id)

  # OpenSpiel (PySpiel) API functions are below. This is the standard set that
  # should be implemented by every perfect-information sequential-move game.

  def mean_field_population(self):
    return self._population

  def _legal_actions(self, player):
    """Returns a list of legal actions for player and MFG nodes."""
    if player == pyspiel.PlayerId.MEAN_FIELD:
      return []
    if player >= 0 and player == self.current_player():
      return list(self._ACTION_TO_MOVE)
    raise ValueError(f"Unexpected player {player}."
                     "Expected a mean field or current player >=0.")

  def chance_outcomes(self) -> List[Tuple[int, float]]:
    """Returns the possible chance outcomes and their probabilities."""
    if self._is_position_init:
      if (self._population is None or
          not 0 <= self._population < self.num_players()):
        raise ValueError(f"Invalid population {self._population}")
      # Make each population start in one corner (this works best when
      # we have 4 populations). When there are more than 4
      # populations, there will be some corners will at least 2
      # populations.
      p = self._population % 4
      initial_position = np.array(
          [p % 2 * (self.size - 1), p // 2 * (self.size - 1)])
      return [(pos_to_merged(initial_position, self.size), 1.)]
    # Uniform distribution over actions.
    return [(action, 1 / len(self._ACTION_TO_MOVE))
            for action in self._ACTION_TO_MOVE]

  def update_pos(self, action):
    """Updates the position of the player given a move action."""
    if action < 0 or action >= len(self._ACTION_TO_MOVE):
      raise ValueError(
          f"The action must be between 0 and {len(self._ACTION_TO_MOVE)}, "
          f"got {action}")
    candidate_pos = self._pos + self._ACTION_TO_MOVE[action]
    if self.geometry == Geometry.TORUS:
      candidate_pos += self.size
      candidate_pos %= self.size
    else:
      assert self.geometry == Geometry.SQUARE, (
          f"Invalid geometry {self.geometry}")
      # Keep the position within the bounds of the square.
      candidate_pos = np.minimum(candidate_pos, self.size - 1)
      candidate_pos = np.maximum(candidate_pos, 0)
    self._pos = candidate_pos

  def _apply_action(self, action):
    """Applies the specified action to the state."""
    if self._population is None:
      raise ValueError(
          "Attempting to perform an action with a population-less state.")
    if self._player_id == pyspiel.PlayerId.MEAN_FIELD:
      raise ValueError(
          "_apply_action should not be called at a MEAN_FIELD state.")
    self._returns += np.array(self.rewards())
    if self._is_position_init:
      self._pos = merged_to_pos(action, self.size)
      self._is_position_init = False
      self._player_id = self._population
    elif self._player_id == pyspiel.PlayerId.CHANCE:
      self.update_pos(action)
      self._t += 1
      self._player_id = pyspiel.PlayerId.MEAN_FIELD
    elif int(self._player_id) >= 0:
      assert self._player_id == self._population, (
          f"Invalid decision player id {self._player_id} "
          f"expected {self._population}")
      self.update_pos(action)
      self._player_id = pyspiel.PlayerId.CHANCE
    else:
      raise ValueError(f"Unexpected state. Player id: {self._player_id}")

  def _action_to_string(self, player, action):
    """Action -> string."""
    del player
    if self.is_chance_node() and self._is_position_init:
      return f"init_position={action}"
    return str(self._ACTION_TO_MOVE[action])

  def distribution_support(self):
    """Returns a list of state string."""
    support = []
    for x in range(self.size):
      for y in range(self.size):
        for population in range(self.num_players()):
          support.append(
              self.state_to_str(
                  np.array([x, y]),
                  self._t,
                  population,
                  player_id=pyspiel.PlayerId.MEAN_FIELD))
    return support

  def get_pos_proba(self, pos: np.ndarray, population: int) -> float:
    """Gets the probability of a pos and population in the current distrib.

    Args:
      pos: 2D position.
      population: Population requested.

    Returns:
      The probability for the provided position and population.
    """
    assert (pos >= 0).all(), pos
    assert (pos < self.size).all(), pos
    assert 0 <= population < self.num_players(), population
    # This logic needs to match the ordering defined in distribution_support().
    index = population + self.num_players() * (pos[1] + self.size * pos[0])
    assert 0 <= index < len(self._distribution.value), (
        f"Invalid index {index} vs dist length: {len(self._distribution.value)}"
        f", population={population}, pos={pos}, state={self}")
    return self._distribution.value[index]

  def update_distribution(self, distribution):
    """This function is central and specific to the logic of the MFG.

    It should only be called when the node is in MEAN_FIELD state.

    Args:
      distribution: List of floats that should contain the probability of each
        state returned by distribution_support().
    """
    expected_dist_size = self.num_states * self.num_players()
    assert len(distribution) == expected_dist_size, (
        f"Unexpected distribution length "
        f"{len(distribution)} != {expected_dist_size}")
    if self._player_id != pyspiel.PlayerId.MEAN_FIELD:
      raise ValueError(
          "update_distribution should only be called at a MEAN_FIELD state.")
    self._distribution = shared_value.SharedValue(distribution)
    self._player_id = self._population

  def is_terminal(self):
    """Returns True if the game is over."""
    return self.t >= self.horizon

  def current_player(self):
    """Returns id of the next player to move, or TERMINAL if game is over."""
    if self.is_terminal():
      return pyspiel.PlayerId.TERMINAL
    return self._player_id

  def rewards(self) -> List[float]:
    """Predator-prey rewards for all populations.

    See section 5.4, paragraph Environment in https://arxiv.org/abs/2103.00623.

    Returns:
      One float per population.
    """
    if int(self._player_id) < 0:
      return [0.] * self.num_players()
    # TODO(author15): Remove this eps once b/191064186 is fixed.
    eps = 1e-25
    densities = np.array([
        self.get_pos_proba(self._pos, population)
        for population in range(self.num_players())
    ],
                         dtype=np.float64)
    rew = -np.log(densities + eps) + np.dot(self.reward_matrix, densities)
    return list(rew)

  def returns(self) -> List[float]:
    """Returns is the sum of all payoffs collected so far."""
    return list(self._returns + np.array(self.rewards()))

  def __str__(self):
    """A string that uniquely identify the current state."""
    return self.state_to_str(
        self._pos, self._t, self._population, player_id=self._player_id)


class Observer:
  """Observer, conforming to the PyObserver interface (see observation.py)."""

  def __init__(self, params, game):
    """Initializes an empty observation tensor."""
    del params

    self.size = game.size
    self.horizon = game.horizon
    # +1 to allow t == horizon.
    self.tensor = np.zeros(2 * self.size + self.horizon + 1, np.float32)
    self.dict = {
        "x": self.tensor[:self.size],
        "y": self.tensor[self.size:self.size * 2],
        "t": self.tensor[self.size * 2:]
    }

  def set_from(self, state: MFGPredatorPreyState, player: int):
    """Updates `tensor` and `dict` to reflect `state` from PoV of `player`."""
    del player
    # We update the observation via the shaped tensor since indexing is more
    # convenient than with the 1-D tensor. Both are views onto the same memory.
    self.tensor.fill(0)
    # state.pos is None for the initial (blank) state, don't set any
    # position bit in that case.
    if state.pos is not None:
      if not (state.pos >= 0).all() or not (state.pos < self.size).all():
        raise ValueError(
            f"Expected {state} positions to be in [0, {self.size})")
      self.dict["x"][state.pos[0]] = 1
      self.dict["y"][state.pos[1]] = 1
    if not 0 <= state.t <= self.horizon:
      raise ValueError(f"Expected {state} time to be in [0, {self.horizon}]")
    self.dict["t"][state.t] = 1

  def string_from(self, state, player):
    """Observation of `state` from the PoV of `player`, as a string."""
    del player
    return str(state)


pyspiel.register_game(_GAME_TYPE, MFGPredatorPreyGame)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for open_spiel.python.algorithms.stackelberg_lp."""

from absl.testing import absltest
from absl.testing import parameterized
import nashpy as nash
import numpy as np

from open_spiel.python.algorithms.stackelberg_lp import solve_stackelberg
from open_spiel.python.egt.utils import game_payoffs_array
import pyspiel

# game instances based on Conitzer & Sandholm'06 paper
game0 = pyspiel.create_matrix_game([[2, 4], [1, 3]], [[1, 0], [0, 1]])
commit_strategy0 = np.array([0.5, 0.5])
commit_value0 = 3.5

game1 = pyspiel.create_matrix_game([[2, 0, 0], [1, 0, 0]],
                                   [[0, 2, 5], [0, -1, -4]])
commit_strategy1 = np.array([1 / 3, 2 / 3])
commit_value1 = 4 / 3


class StackelbergLPTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ("game0", game0, commit_strategy0, commit_value0),
      ("game1", game1, commit_strategy1, commit_value1),
  )
  def test_simple_games(self, game, commit_strategy, commit_value):
    leader_eq_strategy, _, leader_eq_value, _ = solve_stackelberg(game)

    with self.subTest("optimal commitment"):
      np.testing.assert_array_almost_equal(commit_strategy, leader_eq_strategy)
      self.assertAlmostEqual(commit_value, leader_eq_value)

    with self.subTest("Leader-payoff in SSE no less than in NE"):
      p_mat = game_payoffs_array(game)
      nashpy_game = nash.Game(p_mat[0], p_mat[1])
      for eq in nashpy_game.support_enumeration():
        leader_nash_value = eq[0].reshape(1,
                                          -1).dot(p_mat[0]).dot(eq[1].reshape(
                                              -1, 1))
        self.assertGreaterEqual(leader_eq_value, leader_nash_value)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for distribution."""

from absl.testing import absltest
from open_spiel.python import policy
from open_spiel.python.mfg import games  # pylint: disable=unused-import
from open_spiel.python.mfg.algorithms import distribution
import pyspiel


class DistributionTest(absltest.TestCase):

  def test_basic(self):
    game = pyspiel.load_game("python_mfg_crowd_modelling")
    uniform_policy = policy.UniformRandomPolicy(game)
    dist = distribution.DistributionPolicy(game, uniform_policy)
    state = game.new_initial_state().child(0)
    self.assertAlmostEqual(dist.value(state), 1 / game.size)

  def test_state_support_outside_distrib(self):
    game = pyspiel.load_game("mfg_crowd_modelling_2d", {
        "initial_distribution": "[0|0]",
        "initial_distribution_value": "[1.]",
    })
    uniform_policy = policy.UniformRandomPolicy(game)
    _ = distribution.DistributionPolicy(game, uniform_policy)

  def test_multi_pop(self):
    game = pyspiel.load_game("python_mfg_predator_prey")
    self.assertEqual(game.num_players(), 3)
    uniform_policy = policy.UniformRandomPolicy(game)
    dist = distribution.DistributionPolicy(game, uniform_policy)
    for pop in range(3):
      self.assertAlmostEqual(
          dist.value(game.new_initial_state_for_population(pop)), 1.)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""RL agents trained against fixed policy/bot as approximate responses.

This can be used to try to find exploits in policies or bots, as described in
Timbers et al. '20 (https://arxiv.org/abs/2004.09677), but only using RL
directly rather than RL+Search.
"""

from absl import app
from absl import flags
import numpy as np
import tensorflow.compat.v1 as tf

from open_spiel.python import rl_agent
from open_spiel.python import rl_environment
from open_spiel.python import rl_tools
from open_spiel.python.algorithms import dqn
from open_spiel.python.algorithms import random_agent
from open_spiel.python.algorithms import tabular_qlearner

FLAGS = flags.FLAGS

# Training parameters
flags.DEFINE_string("checkpoint_dir", "/tmp/dqn_test",
                    "Directory to save/load the agent models.")
flags.DEFINE_integer(
    "save_every", int(1e4),
    "Episode frequency at which the DQN agent models are saved.")
flags.DEFINE_integer("num_train_episodes", int(1e6),
                     "Number of training episodes.")
flags.DEFINE_integer(
    "eval_every", 1000,
    "Episode frequency at which the DQN agents are evaluated.")
flags.DEFINE_integer("eval_episodes", 1000,
                     "How many episodes to run per eval.")

# DQN model hyper-parameters
flags.DEFINE_list("hidden_layers_sizes", [64, 64, 64],
                  "Number of hidden units in the Q-Network MLP.")
flags.DEFINE_integer("replay_buffer_capacity", int(1e5),
                     "Size of the replay buffer.")
flags.DEFINE_integer("batch_size", 32,
                     "Number of transitions to sample at each learning step.")


# Main algorithm parameters
flags.DEFINE_integer("seed", 0, "Seed to use for everything")
flags.DEFINE_integer("window_size", 30, "Size of window for rolling average")
flags.DEFINE_integer("num_players", 2, "Numebr of players")
flags.DEFINE_string("game", "leduc_poker", "Game string")
flags.DEFINE_string("exploitee", "random", "Exploitee (random | first)")
flags.DEFINE_string("learner", "qlearning", "Learner (qlearning | dqn)")


def eval_against_fixed_bots(env, trained_agents, fixed_agents, num_episodes):
  """Evaluates `trained_agents` against `random_agents` for `num_episodes`."""
  num_players = len(fixed_agents)
  sum_episode_rewards = np.zeros(num_players)
  for player_pos in range(num_players):
    cur_agents = fixed_agents[:]
    cur_agents[player_pos] = trained_agents[player_pos]
    for _ in range(num_episodes):
      time_step = env.reset()
      episode_rewards = 0
      turn_num = 0
      while not time_step.last():
        turn_num += 1
        player_id = time_step.observations["current_player"]
        if env.is_turn_based:
          agent_output = cur_agents[player_id].step(
              time_step, is_evaluation=True)
          action_list = [agent_output.action]
        else:
          agents_output = [
              agent.step(time_step, is_evaluation=True) for agent in cur_agents
          ]
          action_list = [agent_output.action for agent_output in agents_output]
        time_step = env.step(action_list)
        episode_rewards += time_step.rewards[player_pos]
      sum_episode_rewards[player_pos] += episode_rewards
  return sum_episode_rewards / num_episodes


def create_training_agents(num_players, sess, num_actions, info_state_size,
                           hidden_layers_sizes):
  """Create the agents we want to use for learning."""
  if FLAGS.learner == "qlearning":
    # pylint: disable=g-complex-comprehension
    return [
        tabular_qlearner.QLearner(
            player_id=idx,
            num_actions=num_actions,
            # step_size=0.02,
            step_size=0.1,
            # epsilon_schedule=rl_tools.ConstantSchedule(0.5),
            epsilon_schedule=rl_tools.LinearSchedule(0.5, 0.2, 1000000),
            discount_factor=0.99) for idx in range(num_players)
    ]
  elif FLAGS.learner == "dqn":
    # pylint: disable=g-complex-comprehension
    return [
        dqn.DQN(
            session=sess,
            player_id=idx,
            state_representation_size=info_state_size,
            num_actions=num_actions,
            discount_factor=0.99,
            epsilon_start=0.5,
            epsilon_end=0.1,
            hidden_layers_sizes=hidden_layers_sizes,
            replay_buffer_capacity=FLAGS.replay_buffer_capacity,
            batch_size=FLAGS.batch_size) for idx in range(num_players)
    ]
  else:
    raise RuntimeError("Unknown learner")


class FirstActionAgent(rl_agent.AbstractAgent):
  """An example agent class."""

  def __init__(self, player_id, num_actions, name="first_action_agent"):
    assert num_actions > 0
    self._player_id = player_id
    self._num_actions = num_actions

  def step(self, time_step, is_evaluation=False):
    # If it is the end of the episode, don't select an action.
    if time_step.last():
      return

    # Pick the first legal action.
    cur_legal_actions = time_step.observations["legal_actions"][self._player_id]
    action = cur_legal_actions[0]
    probs = np.zeros(self._num_actions)
    probs[action] = 1.0

    return rl_agent.StepOutput(action=action, probs=probs)


class RollingAverage(object):
  """Class to store a rolling average."""

  def __init__(self, size=100):
    self._size = size
    self._values = np.array([0] * self._size, dtype=np.float64)
    self._index = 0
    self._total_additions = 0

  def add(self, value):
    self._values[self._index] = value
    self._total_additions += 1
    self._index = (self._index + 1) % self._size

  def mean(self):
    n = min(self._size, self._total_additions)
    if n == 0:
      return 0
    return self._values.sum() / n


def main(_):
  np.random.seed(FLAGS.seed)
  tf.random.set_random_seed(FLAGS.seed)

  num_players = FLAGS.num_players

  env = rl_environment.Environment(FLAGS.game, include_full_state=True)
  info_state_size = env.observation_spec()["info_state"][0]
  num_actions = env.action_spec()["num_actions"]

  # Exploitee agents
  if FLAGS.exploitee == "first":
    exploitee_agents = [
        FirstActionAgent(idx, num_actions) for idx in range(num_players)
    ]
  elif FLAGS.exploitee == "random":
    exploitee_agents = [
        random_agent.RandomAgent(player_id=idx, num_actions=num_actions)
        # FirstActionAgent(player_id=idx, num_actions=num_actions)
        for idx in range(num_players)
    ]
  else:
    raise RuntimeError("Unknown exploitee")

  rolling_averager = RollingAverage(FLAGS.window_size)
  rolling_averager_p0 = RollingAverage(FLAGS.window_size)
  rolling_averager_p1 = RollingAverage(FLAGS.window_size)
  rolling_value = 0
  total_value = 0
  total_value_n = 0

  with tf.Session() as sess:
    hidden_layers_sizes = [int(l) for l in FLAGS.hidden_layers_sizes]
    # pylint: disable=g-complex-comprehension
    learning_agents = create_training_agents(num_players, sess, num_actions,
                                             info_state_size,
                                             hidden_layers_sizes)
    sess.run(tf.global_variables_initializer())

    print("Starting...")

    for ep in range(FLAGS.num_train_episodes):
      if (ep + 1) % FLAGS.eval_every == 0:
        r_mean = eval_against_fixed_bots(env, learning_agents, exploitee_agents,
                                         FLAGS.eval_episodes)
        value = r_mean[0] + r_mean[1]
        rolling_averager.add(value)
        rolling_averager_p0.add(r_mean[0])
        rolling_averager_p1.add(r_mean[1])
        rolling_value = rolling_averager.mean()
        rolling_value_p0 = rolling_averager_p0.mean()
        rolling_value_p1 = rolling_averager_p1.mean()
        total_value += value
        total_value_n += 1
        avg_value = total_value / total_value_n
        print(("[{}] Mean episode rewards {}, value: {}, " +
               "rval: {} (p0/p1: {} / {}), aval: {}").format(
                   ep + 1, r_mean, value, rolling_value, rolling_value_p0,
                   rolling_value_p1, avg_value))

      agents_round1 = [learning_agents[0], exploitee_agents[1]]
      agents_round2 = [exploitee_agents[0], learning_agents[1]]

      for agents in [agents_round1, agents_round2]:
        time_step = env.reset()
        while not time_step.last():
          player_id = time_step.observations["current_player"]
          if env.is_turn_based:
            agent_output = agents[player_id].step(time_step)
            action_list = [agent_output.action]
          else:
            agents_output = [agent.step(time_step) for agent in agents]
            action_list = [
                agent_output.action for agent_output in agents_output
            ]
          time_step = env.step(action_list)

        # Episode is over, step all agents with final info state.
        for agent in agents:
          agent.step(time_step)


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Various masked_softmax implementations, both in numpy and tensorflow."""

import numpy as np
import tensorflow.compat.v1 as tf

# Temporarily disable TF2 behavior until the code is updated.
tf.disable_v2_behavior()


def tf_masked_softmax(logits, legal_actions_mask):
  """Returns the softmax over the valid actions defined by `legal_actions_mask`.

  Args:
    logits: A tensor [..., num_actions] (e.g. [num_actions] or [B, num_actions])
      representing the logits to mask.
    legal_actions_mask: The legal action mask, same shape as logits. 1 means
      it's a legal action, 0 means it's illegal. If can be a tensorflow or numpy
      tensor.
  """
  # This will raise a warning as we are taking the log of 0, which sets the 0
  # values to -inf. However, this is fine, as we then apply tf.exp, which sets
  # tf.exp(-inf) to 0. e.g. if we have logits [5, 3, 1], with legal_mask
  # [0, 1, 1], then masked_logits == [-inf, 3, 1], so we subtract the max to
  # get [-inf, 0, -2], and apply tf.exp to get [0, 1, e^-2].
  legal_actions_mask = tf.cast(legal_actions_mask, dtype=logits.dtype)
  masked_logits = logits + tf.log(legal_actions_mask)
  max_logit = tf.reduce_max(masked_logits, axis=-1, keepdims=True)
  exp_logit = tf.exp(masked_logits - max_logit)
  return exp_logit / tf.reduce_sum(exp_logit, axis=-1, keepdims=True)


def np_masked_softmax(logits, legal_actions_mask):
  """Returns the softmax over the valid actions defined by `legal_actions_mask`.

  Args:
    logits: A tensor [..., num_actions] (e.g. [num_actions] or [B, num_actions])
      representing the logits to mask.
    legal_actions_mask: The legal action mask, same shape as logits. 1 means
      it's a legal action, 0 means it's illegal.
  """
  masked_logits = logits + np.log(legal_actions_mask)
  max_logit = np.amax(masked_logits, axis=-1, keepdims=True)
  exp_logit = np.exp(masked_logits - max_logit)
  return exp_logit / np.sum(exp_logit, axis=-1, keepdims=True)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for RNaD algorithm under open_spiel."""

import pickle

from absl.testing import absltest
import jax
import numpy as np

from open_spiel.python.algorithms.rnad import rnad

# TODO(author18): test the losses and jax ops


class RNADTest(absltest.TestCase):

  def test_run_kuhn(self):
    solver = rnad.RNaDSolver(rnad.RNaDConfig(game_name="kuhn_poker"))
    for _ in range(10):
      solver.step()

  def test_serialization(self):
    solver = rnad.RNaDSolver(rnad.RNaDConfig(game_name="kuhn_poker"))
    solver.step()

    state_bytes = pickle.dumps(solver)
    solver2 = pickle.loads(state_bytes)

    self.assertEqual(solver.config, solver2.config)
    np.testing.assert_equal(
        jax.device_get(solver.params), jax.device_get(solver2.params))

    # TODO(author16): figure out the last bits of the non-determinism
    #             and reenable the checks below.
    # Now run both solvers for the same number of steps and verify
    # they behave in exactly the same way.
    # for _ in range(10):
    #   solver.step()
    #   solver2.step()
    # np.testing.assert_equal(
    #     jax.device_get(solver.params), jax.device_get(solver2.params))


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for meta CFR Algorithm."""

from absl import flags
from absl.testing import absltest
from absl.testing import parameterized
import haiku as hk
import jax
import mock
import numpy as np
import optax

from open_spiel.python.examples.meta_cfr.sequential_games import meta_learning
from open_spiel.python.examples.meta_cfr.sequential_games import models
from open_spiel.python.examples.meta_cfr.sequential_games import openspiel_api

FLAGS = flags.FLAGS


def meta_cfr_agent(game_name='kuhn_poker'):
  return meta_learning.MetaCFRRegretAgent(
      training_epochs=1,
      meta_learner_training_epochs=1,
      game_name=game_name,
      game_config={'players': 2},
      perturbation=False,
      seed=0,
      model_type='MLP',
      best_response=True)


class MetaLearningTest(parameterized.TestCase):

  def setup_optimizer(self, num_actions, num_infostates):
    if FLAGS.use_infostate_representation:
      dummy_input = np.zeros(
          shape=[FLAGS.batch_size, 1, num_actions + num_infostates])
    else:
      dummy_input = np.zeros(shape=[FLAGS.batch_size, 1, num_actions])

    def mlp_forward(dummy_input):
      mlp = hk.nets.MLP([10, num_actions])
      return mlp(dummy_input)
    forward = hk.transform(mlp_forward)

    rng_seq = jax.random.PRNGKey(10)
    params = forward.init(rng_seq, dummy_input)
    lr_scheduler_fn = optax.polynomial_schedule(
        init_value=0.2, end_value=0.0001, power=1., transition_steps=100)
    opt_init, opt_update = optax.chain(
        optax.scale_by_adam(), optax.scale_by_schedule(lr_scheduler_fn),
        optax.scale(-0.2))
    net_apply = forward.apply
    opt_state = opt_init(params)
    return params, net_apply, opt_state, opt_update

  @parameterized.named_parameters(('kuhn_poker_game', 'kuhn_poker'),
                                  ('leduc_poker_game', 'leduc_poker'))
  def test_worldstate_initialization(self, game_name):
    self._world_state = openspiel_api.WorldState(
        game_name, {'players': 2}, perturbation=False, random_seed=0)
    self._all_actions = self._world_state.get_distinct_actions()
    self.assertNotEmpty(self._all_actions,
                        'Number of distinct actions should be greater that 0.')

  @parameterized.named_parameters(('kuhn_poker_game', 'kuhn_poker'),
                                  ('leduc_poker_game', 'leduc_poker'))
  def test_meta_cfr_agent_initialization(self, game_name):
    with mock.patch.object(meta_learning.MetaCFRRegretAgent,
                           'get_num_infostates') as mock_get_num_infostates:
      mock_get_num_infostates.return_value = (mock.MagicMock(),
                                              mock.MagicMock())
      meta_learning.MetaCFRRegretAgent(
          training_epochs=1,
          meta_learner_training_epochs=1,
          game_name=game_name,
          game_config={'players': 2},
          perturbation=False,
          seed=0,
          model_type='MLP',
          best_response=True)
    mock_get_num_infostates.assert_called_once_with()

  @parameterized.named_parameters(('kuhn_poker_game', 'kuhn_poker'),
                                  ('leduc_poker_game', 'leduc_poker'))
  def test_meta_learning_training(self, game_name):
    agent = meta_learning.MetaCFRRegretAgent(
        training_epochs=1,
        meta_learner_training_epochs=1,
        game_name=game_name,
        game_config={'players': 2},
        perturbation=False,
        seed=0,
        model_type=models.ModelType.MLP.value,
        best_response=True)
    num_infostates, _ = agent.get_num_infostates()
    num_actions = len(agent._all_actions)
    params, net_apply, opt_state, opt_update = self.setup_optimizer(
        num_actions, num_infostates)
    agent.training_optimizer()
    agent.optimizer.net_apply = net_apply
    agent.optimizer.opt_state = opt_state
    agent.optimizer.net_params = params
    agent.optimizer.opt_update = opt_update

    world_state = openspiel_api.WorldState(
        game_name, {'players': 2}, perturbation=False, random_seed=0)
    best_response_val_player_2 = agent.next_policy(world_state)
    self.assertGreater(best_response_val_player_2[-1], 0)

if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Counterfactual Regret Minimization."""

import copy
import enum
from typing import List, Tuple

from open_spiel.python.examples.meta_cfr.sequential_games.typing import GameTree
from open_spiel.python.examples.meta_cfr.sequential_games.typing import HistoryNode
from open_spiel.python.examples.meta_cfr.sequential_games.typing import InfostateMapping
from open_spiel.python.examples.meta_cfr.sequential_games.typing import InfostateNode


class Players(enum.IntEnum):
  CHANCE_PLAYER = 0
  PLAYER_1 = 1
  PLAYER_2 = 2


def compute_reach_probabilities(
    history_tree_node: HistoryNode,
    all_infostates_map: List[InfostateMapping]) -> None:
  """Computes reach probabilities for game tree information states.

  This function initializes counterfactual_reach_prob and player_reach_prob for
  all information states in the game tree, and then these values will be
  calculated in compute_reach_probability_dfs.

  Args:
    history_tree_node: Game tree HistoryTreeNode which is the root of the game
      tree.
    all_infostates_map: List of dictionaries (mapping from information state
      string representation to information state object) for each players
      (including chance player). This list will be empty when this function is
      called fot the first time and it'll be population during DFS tree
      traversal.
  """

  for infostate in (list(all_infostates_map[Players.PLAYER_1].values()) +
                    list(all_infostates_map[Players.PLAYER_2].values())):
    infostate.counterfactual_reach_prob = 0.
    infostate.player_reach_prob = 0.
  compute_reach_probability_dfs(history_tree_node, all_infostates_map)


def compute_reach_probability_dfs(
    history_tree_node: HistoryNode,
    all_infostates_map: List[InfostateMapping]) -> None:
  """Calculate reach probability values in dfs tree.

  This function is initially called by compute_reach_probabilities and it
  computes reach probabilities for all information state nodes in the tree by
  traversing the tree using DFS.

  Args:
    history_tree_node: Game tree HistoryTreeNode which is the root of the game
      tree.
    all_infostates_map: List of dictionaries (mapping from information state
      string representation to information state object) for each players
      (including chance player). This list will be empty when this function is
      called fot the first time and it'll be population during DFS tree
      traversal.
  """

  world_state = history_tree_node.world_state
  infostate_p1 = all_infostates_map[Players.PLAYER_1][
      world_state.get_infostate_string(Players.PLAYER_1)]
  infostate_p2 = all_infostates_map[Players.PLAYER_2][
      world_state.get_infostate_string(Players.PLAYER_2)]
  infostate_p1.counterfactual_reach_prob += history_tree_node.reach_probs[
      0] * history_tree_node.reach_probs[Players.PLAYER_2]
  infostate_p2.counterfactual_reach_prob += history_tree_node.reach_probs[
      0] * history_tree_node.reach_probs[Players.PLAYER_1]

  if infostate_p1.player_reach_prob != 0.:
    assert (infostate_p1.player_reach_prob == history_tree_node.reach_probs[
        Players.PLAYER_1])

  if infostate_p2.player_reach_prob != 0.:
    assert (infostate_p2.player_reach_prob == history_tree_node.reach_probs[
        Players.PLAYER_2])

  infostate_p1.player_reach_prob = history_tree_node.reach_probs[
      Players.PLAYER_1]
  infostate_p2.player_reach_prob = history_tree_node.reach_probs[
      Players.PLAYER_2]

  policy_p1 = infostate_p1.policy
  policy_p2 = infostate_p2.policy
  policy_chance = world_state.chance_policy
  actions_chance, actions_p1, actions_p2 = world_state.get_actions()
  for action_chance in actions_chance:
    for action_p1 in actions_p1:
      for action_p2 in actions_p2:
        history_tree_node.action_probs[(
            action_chance, action_p1, action_p2)] = policy_chance[
                action_chance] * policy_p1[action_p1] * policy_p2[action_p2]
        child_node = history_tree_node.get_child(
            (action_chance, action_p1, action_p2))
        child_node.reach_probs[
            Players.CHANCE_PLAYER] = history_tree_node.reach_probs[
                Players.CHANCE_PLAYER] * policy_chance[action_chance]
        child_node.reach_probs[
            Players.PLAYER_1] = history_tree_node.reach_probs[
                Players.PLAYER_1] * policy_p1[action_p1]
        child_node.reach_probs[
            Players.PLAYER_2] = history_tree_node.reach_probs[
                Players.PLAYER_2] * policy_p2[action_p2]
        compute_reach_probability_dfs(child_node, all_infostates_map)


def _get_opponent(player: int) -> int:
  return -1 * player + 3


def compute_best_response_values(infostate: InfostateNode) -> float:
  """Returns best response value for an infostate.

  Args:
    infostate: Information state.

  Returns:
    Best response value, which is the maximum action value chosen among all
    actions values of possible actions from infostate. If information state is a
    terminal node in the game tree, this value is calculated from history nodes
    reach probability for player and opponent, and game utility of terminal
    node. If infostate is not terminal, this value will be calculated in a
    recursive way.
  """
  if infostate.is_terminal():
    terminal_utility = 0
    for history_node in infostate.history_nodes:
      terminal_utility += history_node.reach_probs[
          0] * history_node.reach_probs[_get_opponent(
              infostate.player)] * history_node.world_state.get_utility(
                  infostate.player)
    return terminal_utility
  action_values = {action: 0 for action in infostate.get_actions()}
  infostate_actions = infostate.get_actions()
  for action in infostate_actions:
    action_values[action] = 0
    for child in infostate.children[action].values():
      action_values[action] += compute_best_response_values(child)
  return max(action_values.values())


def compute_best_response_policy(infostate: InfostateNode) -> float:
  """Calculate best response policy and returns best response value of infostate.

  Args:
    infostate: Information state.

  Returns:
    Best response value similar to what compute_best_response_values returns.
  """
  if infostate.is_terminal():
    terminal_utility = 0
    for history_node in infostate.history_nodes:
      terminal_utility += history_node.reach_probs[
          0] * history_node.reach_probs[_get_opponent(
              infostate.player)] * history_node.world_state.get_utility(
                  infostate.player)
    return terminal_utility
  action_values = {action: 0 for action in infostate.get_actions()}
  infostate_actions = infostate.get_actions()
  for action in infostate_actions:
    action_values[action] = 0
    for child in infostate.children[action].values():
      action_values[action] += compute_best_response_policy(child)

  infostate.policy = {action: 0 for action in infostate.get_actions()}
  max_action_value = max(action_values.values())
  for action in infostate_actions:
    if action_values[action] == max_action_value:
      infostate.policy[action] = 1
      break
  return max_action_value


def compute_counterfactual_values(infostate: InfostateNode) -> float:
  """Returns cfr value for an infostate.

  Args:
    infostate: Information state.

  Returns:
    Counterfactual value for infostate. This value is calculated from action
    value and policy of all legal actions of infostate information state.
  """
  if infostate.is_terminal():
    terminal_utility = 0
    for history_node in infostate.history_nodes:
      terminal_utility += history_node.reach_probs[
          0] * history_node.reach_probs[_get_opponent(
              infostate.player)] * history_node.world_state.get_utility(
                  infostate.player)
    return terminal_utility
  infostate_actions = infostate.get_actions()
  action_values = {action: 0 for action in infostate_actions}
  for action in infostate_actions:
    for child in infostate.children[action].values():
      action_values[action] += compute_counterfactual_values(child)
  infostate.counterfactual_action_values = action_values
  counterfactual_value = 0
  for action in infostate_actions:
    counterfactual_value += infostate.policy[action] * action_values[action]
  infostate.counterfactual_value = counterfactual_value
  return counterfactual_value


def update_regrets(infostates: List[InfostateNode]) -> None:
  """Updates regret value for each infostate in infostates.

  Args:
    infostates: List of information states
  """
  for infostate in infostates:
    for action in infostate.get_actions():
      current_regret = infostate.counterfactual_action_values[
          action] - infostate.counterfactual_value
      infostate.regret[action] += current_regret


def compute_next_policy(infostates: List[InfostateNode],
                        cfr_plus: bool = False) -> None:
  """Computes policy of next iteration for each infostate in infostates.

  Args:
    infostates: List of information states.
    cfr_plus: A flag which specifies if we update policy according to CFR or
      CFR-plus algorithm. True if we use CFR-plus, otherwise we use CFR.
  """
  for infostate in infostates:
    infostate_actions = infostate.get_actions()
    if cfr_plus:
      for action in infostate_actions:
        infostate.regret[action] = max(infostate.regret[action], 0.0)

    positive_regret_sum = 0
    for action in infostate_actions:
      if infostate.regret[action] > 0:
        positive_regret_sum += infostate.regret[action]

    actions_count = len(infostate_actions)
    next_policy = {a: 1.0 / actions_count for a in infostate_actions}

    if positive_regret_sum > 0:
      for action in infostate_actions:
        next_policy[action] = max(infostate.regret[action],
                                  0) / positive_regret_sum
    infostate.policy = next_policy


def cumulate_average_policy(infostates: List[InfostateNode],
                            weight: int = 1) -> None:
  """Cumulates policy values of each infostate in infostates.

  For each infostate, we update average policy and the sum of weighted average
  policy.

  Args:
    infostates: List of information states.
    weight: The weight we use to update policy and sum of weighted average
      policy. For CFR algorithm, weight is 1.
  """
  for infostate in infostates:
    for action in infostate.get_actions():
      infostate.average_policy[
          action] += infostate.player_reach_prob * infostate.policy[
              action] * weight
    infostate.average_policy_weight_sum += infostate.player_reach_prob * weight


def normalize_average_policy(infostates) -> None:
  """Updates infostate policy by normalizing average policy.

  Args:
    infostates: List of information states that their policies will be updated.
  """
  for infostate in infostates:
    for action in infostate.get_actions():
      infostate.policy[action] = infostate.average_policy[
          action] / infostate.average_policy_weight_sum


def best_response_counterfactual_regret_minimization_iteration(
    history_tree_node: HistoryNode,
    infostate_nodes: List[InfostateNode],
    all_infostates_map: List[InfostateMapping]) -> None:
  """Calculates CFRBR values.

  Args:
    history_tree_node: Game tree HistoryTreeNode which is the root of the game
      tree.
    infostate_nodes: List of all information state nodes.
    all_infostates_map: List of dictionaries (mapping from information state
      string representation to information state object) for each players
      (including chance player). This list will be empty when this function is
      called fot the first time and it'll be population during DFS tree
      traversal.
  """
  compute_next_policy(list(all_infostates_map[Players.PLAYER_1].values()))

  compute_reach_probabilities(history_tree_node, all_infostates_map)
  cumulate_average_policy(list(all_infostates_map[Players.PLAYER_1].values()))

  compute_best_response_policy(infostate_nodes[Players.PLAYER_2])
  compute_reach_probabilities(history_tree_node, all_infostates_map)
  compute_counterfactual_values(infostate_nodes[Players.PLAYER_1])

  update_regrets(list(all_infostates_map[Players.PLAYER_1].values()))


def counterfactual_regret_minimization_iteration(
    cfr_game_tree: GameTree,
    alternating_updates: bool,
    cfr_plus: bool,
    weight: int = 1) -> None:
  """Performs one iteration of CFR or CFR-plus.

  Args:
    cfr_game_tree: Game tree for an imperfect information game. This game tree
      is game tree of an openspiel game.
    alternating_updates: Boolean flag to do alternative update for players
      policies or not. If True, alternative updates will be performed (meaning
      we first calculate average policy, counterfactual values, regrets and next
      policy for player 1 first and then calculate all of these for player 2),
      otherwise both players average policies, counterfactual values and regrets
      will be updated right after each other (meaning, for example we calculate
      next_policy of player 1, and then next policy of player 2. Then, we
      calculate average policy for player 1 and then average policy for player
      2, and so on).
    cfr_plus: Boolean flag indicating if we perform CFR algorithm or CFR-plus.
      If True, we perform CFR-plus algorithm, otherwise we perform CFR
      algorithm.
    weight: The weight we use to update policy and sum of weighted average
      policy.
  """
  if alternating_updates:
    compute_reach_probabilities(cfr_game_tree.first_history_node,
                                cfr_game_tree.all_infostates_map)
    cumulate_average_policy(
        list(cfr_game_tree.all_infostates_map[Players.PLAYER_1].values()),
        weight)
    compute_counterfactual_values(
        cfr_game_tree.infostate_nodes[Players.PLAYER_1])
    update_regrets(
        list(cfr_game_tree.all_infostates_map[Players.PLAYER_1].values()))
    compute_next_policy(
        list(cfr_game_tree.all_infostates_map[Players.PLAYER_1].values()),
        cfr_plus)

    compute_reach_probabilities(cfr_game_tree.first_history_node,
                                cfr_game_tree.all_infostates_map)
    cumulate_average_policy(
        list(cfr_game_tree.all_infostates_map[Players.PLAYER_2].values()),
        weight)
    compute_counterfactual_values(
        cfr_game_tree.infostate_nodes[Players.PLAYER_2])
    update_regrets(
        list(cfr_game_tree.all_infostates_map[Players.PLAYER_2].values()))
    compute_next_policy(
        list(cfr_game_tree.all_infostates_map[Players.PLAYER_2].values()),
        cfr_plus)
  else:
    compute_next_policy(
        list(cfr_game_tree.all_infostates_map[Players.PLAYER_1].values()),
        cfr_plus)
    compute_next_policy(
        list(cfr_game_tree.all_infostates_map[Players.PLAYER_2].values()),
        cfr_plus)

    compute_reach_probabilities(cfr_game_tree.first_history_node,
                                cfr_game_tree.all_infostates_map)
    cumulate_average_policy(
        list(cfr_game_tree.all_infostates_map[Players.PLAYER_1].values()),
        weight)
    cumulate_average_policy(
        list(cfr_game_tree.all_infostates_map[Players.PLAYER_2].values()),
        weight)

    compute_counterfactual_values(
        cfr_game_tree.infostate_nodes[Players.PLAYER_1])
    compute_counterfactual_values(
        cfr_game_tree.infostate_nodes[Players.PLAYER_2])

    update_regrets(
        list(cfr_game_tree.all_infostates_map[Players.PLAYER_1].values()))
    update_regrets(
        list(cfr_game_tree.all_infostates_map[Players.PLAYER_2].values()))


def compute_cfr_plus_values(cfr_game_tree: GameTree,
                            steps: int) -> Tuple[List[float], List[float]]:
  """Performs CFR-plus algorithm for a given number of steps.

  Args:
    cfr_game_tree: Game tree for an imperfect information game. This game tree
      is game tree of an openspiel game.
    steps: Number of CFR-plus steps.

  Returns:
    best_response_values_p1: List of best response values for player 1. The
    length of this list is equal to the number of steps.
    best_response_values_p2: List of best response values for player 2. The
    length of this list is equal to the number of steps.
  """
  best_response_values_p1 = []
  best_response_values_p2 = []
  for i in range(steps):
    counterfactual_regret_minimization_iteration(
        cfr_game_tree=cfr_game_tree,
        alternating_updates=True,
        cfr_plus=True,
        weight=i + 1)

    game_tree_copy = copy.deepcopy(cfr_game_tree)
    normalize_average_policy(
        game_tree_copy.all_infostates_map[Players.PLAYER_1].values())
    normalize_average_policy(
        game_tree_copy.all_infostates_map[Players.PLAYER_2].values())
    compute_reach_probabilities(game_tree_copy.first_history_node,
                                game_tree_copy.all_infostates_map)

    best_response_values_p1.append(
        compute_best_response_values(
            game_tree_copy.infostate_nodes[Players.PLAYER_1]))
    best_response_values_p2.append(
        compute_best_response_values(
            game_tree_copy.infostate_nodes[Players.PLAYER_2]))

  return best_response_values_p1, best_response_values_p2


def compute_cfr_values(cfr_game_tree: GameTree,
                       steps: int) -> Tuple[List[float], List[float]]:
  """Performs CFR algorithm for a given number of steps.

  Args:
    cfr_game_tree: Game tree for an imperfect information game. This game tree
      is game tree of an openspiel game.
    steps: Number of CFR-plus steps.

  Returns:
    best_response_values_p1: List of best response values for player 1. The
    length of this list is equal to the number of steps.
    best_response_values_p2: List of best response values for player 2. The
    length of this list is equal to the number of steps.
  """
  best_response_values_p1 = []
  best_response_values_p2 = []
  for _ in range(steps):
    counterfactual_regret_minimization_iteration(
        cfr_game_tree=cfr_game_tree, alternating_updates=False, cfr_plus=False)

    normalize_average_policy(
        cfr_game_tree.all_infostates_map[Players.PLAYER_1].values())
    normalize_average_policy(
        cfr_game_tree.all_infostates_map[Players.PLAYER_2].values())
    compute_reach_probabilities(cfr_game_tree.first_history_node,
                                cfr_game_tree.all_infostates_map)
    best_response_values_p1.append(
        compute_best_response_values(
            cfr_game_tree.infostate_nodes[Players.PLAYER_1]))
    best_response_values_p2.append(
        compute_best_response_values(
            cfr_game_tree.infostate_nodes[Players.PLAYER_2]))

  return best_response_values_p1, best_response_values_p2



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
r"""Two BlueChip bridge bots bid with each other.

The bot_cmd FLAG should contain a command-line to launch an external bot, e.g.
`Wbridge5 Autoconnect {port}`.

"""
# pylint: enable=line-too-long

import re
import socket
import subprocess

from absl import app
from absl import flags
import numpy as np

from open_spiel.python.bots import bluechip_bridge_uncontested_bidding
import pyspiel

FLAGS = flags.FLAGS
flags.DEFINE_float("timeout_secs", 60, "Seconds to wait for bot to respond")
flags.DEFINE_integer("rng_seed", 1234, "Seed to use to generate hands")
flags.DEFINE_integer("num_deals", 10, "How many deals to play")
flags.DEFINE_string(
    "bot_cmd", None,
    "Command to launch the external bot; must include {port} which will be "
    "replaced by the port number to attach to.")


def _run_once(state, bots):
  """Plays bots with each other, returns terminal utility for each player."""
  for bot in bots:
    bot.restart_at(state)
  while not state.is_terminal():
    if state.is_chance_node():
      outcomes, probs = zip(*state.chance_outcomes())
      state.apply_action(np.random.choice(outcomes, p=probs))
    else:
      state.apply_action(bots[state.current_player()].step(state)[1])
  return state


def main(argv):
  if len(argv) > 1:
    raise app.UsageError("Too many command-line arguments.")
  game = pyspiel.load_game("bridge_uncontested_bidding", {
      "relative_scoring": True,
      "rng_seed": FLAGS.rng_seed,
  })
  bots = [
      bluechip_bridge_uncontested_bidding.BlueChipBridgeBot(
          game, 0, _WBridge5Client(FLAGS.bot_cmd)),
      bluechip_bridge_uncontested_bidding.BlueChipBridgeBot(
          game, 1, _WBridge5Client(FLAGS.bot_cmd)),
  ]
  results = []

  for i_deal in range(FLAGS.num_deals):
    state = _run_once(game.new_initial_state(), bots)
    print("Deal #{}; final state:\n{}".format(i_deal, state))
    results.append(state.returns())

  stats = np.array(results)
  mean = np.mean(stats, axis=0)
  stderr = np.std(stats, axis=0, ddof=1) / np.sqrt(FLAGS.num_deals)
  print(u"Absolute score: {:+.1f}\u00b1{:.1f}".format(mean[0], stderr[0]))
  print(u"Relative score: {:+.1f}\u00b1{:.1f}".format(mean[1], stderr[1]))


class _WBridge5Client(object):
  """Manages the connection to a WBridge5 bot."""

  def __init__(self, command):
    self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    self.sock.bind(("", 0))
    self.port = self.sock.getsockname()[1]
    self.sock.listen(1)
    self.process = None
    self.command = command.format(port=self.port)

  def start(self):
    if self.process is not None:
      self.process.kill()
    self.process = subprocess.Popen(self.command.split(" "))
    self.conn, self.addr = self.sock.accept()

  def read_line(self):
    line = ""
    while True:
      self.conn.settimeout(FLAGS.timeout_secs)
      data = self.conn.recv(1024)
      if not data:
        raise EOFError("Connection closed")
      line += data.decode("ascii")
      if line.endswith("\n"):
        return re.sub(r"\s+", " ", line).strip()

  def send_line(self, line):
    self.conn.send((line + "\r\n").encode("ascii"))


if __name__ == "__main__":
  app.run(main)
